# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from http import HTTPStatus
import pecan
import re
import wsme
from wsme import types as wtypes

from oslo_log import log
from oslo_utils import uuidutils

from cyborg import api
from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers import types
from cyborg.api.controllers.v2 import versions
from cyborg.api import expose
from cyborg.common import authorize_wsgi
from cyborg.common import constants
from cyborg.common import exception
from cyborg.common.i18n import _
from cyborg import objects
LOG = log.getLogger(__name__)

"""
The device profile object and db table has a profile_json field, which has
its own version apart from the device profile groups field. The reasoning
behind that was this structure may evolve more rapidly. Since then the
feedback has been to manage this with the API version itself, preferably
with microversions, rather than use a second version.

One problem with that is that we have to decide on a suitable db
representation for device profile groups, which form an array of
string pairs. The Cyborg community wishes to keep the number of
tables small and manageable.

As of now, the db layer for device profiles still uses the profile_json
field. But the API layer returns the device profile as it should be.
The objects layer does the conversion.
"""


class DeviceProfile(base.APIBase):
    """API representation of a device profile.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation of
    a device profile. See module notes above.
    """

    """The UUID of the device profile"""
    uuid = types.uuid

    """The name of the device profile"""
    name = wtypes.text

    """The description of the device profile"""
    description = wtypes.text

    """The groups of the device profile"""
    groups = [types.jsontype]

    created_at = wtypes.datetime.datetime
    updated_at = wtypes.datetime.datetime

    """A list containing a self link"""
    links = wsme.wsattr([link.Link], readonly=True)

    def __init__(self, **kwargs):
        super(DeviceProfile, self).__init__(**kwargs)
        self.fields = []
        for field in objects.DeviceProfile.fields:
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

    @classmethod
    def convert_with_links(cls, obj_devprof):
        api_devprof = cls(**obj_devprof.as_dict())
        api_devprof.links = [
            link.Link.make_link('self', pecan.request.public_url,
                                'device_profiles', api_devprof.uuid)
            ]
        return api_devprof

    def get_device_profile(self, obj_devprof):
        api_obj = {}
        for field in ['name', 'description', 'uuid', 'groups']:
            api_obj[field] = obj_devprof[field]
        for field in ['created_at', 'updated_at']:
            api_obj[field] = str(obj_devprof[field])
        api_obj['links'] = [
            link.Link.make_link_dict('device_profiles', api_obj['uuid'])
            ]
        return api_obj


class DeviceProfileCollection(DeviceProfile):
    """API representation of a collection of device profiles."""

    """A list containing device profile objects"""
    device_profiles = [DeviceProfile]

    @classmethod
    def convert_with_links(cls, obj_devprofs):
        collection = cls()
        collection.device_profiles = [
            DeviceProfile.convert_with_links(obj_devprof)
            for obj_devprof in obj_devprofs]
        return collection

    def get_device_profiles(self, obj_devprofs):
        api_obj_devprofs = [
            self.get_device_profile(obj_devprof)
            for obj_devprof in obj_devprofs]
        return api_obj_devprofs


class DeviceProfilesController(base.CyborgController,
                               DeviceProfileCollection):
    """REST controller for Device Profiles."""

    @authorize_wsgi.authorize_wsgi("cyborg:device_profile", "create", False)
    @expose.expose(DeviceProfile, body=types.jsontype,
                   status_code=HTTPStatus.CREATED)
    def post(self, req_devprof_list):
        """Create one or more device_profiles.

        NOTE: Only one device profile supported in Train.

        :param devprof: a list of device_profiles.
         [{ "name": <string>,
           "groups": [ {"key1: "value1", "key2": "value2"} ]
           "uuid": <uuid> # optional
           "description": <description> # optional
         }]
         :returns: The list of created device profiles
        """
        # TODO(Sundar) Support more than one devprof per request, if needed

        LOG.info("[device_profiles] POST request = (%s)", req_devprof_list)
        if len(req_devprof_list) != 1:
            raise exception.InvalidParameterValue(
                err="Only one device profile allowed "
                    "per POST request for now.")
        req_devprof = req_devprof_list[0]
        self._validate_post_request(req_devprof)

        context = pecan.request.context
        obj_devprof = objects.DeviceProfile(context, **req_devprof)

        new_devprof = pecan.request.conductor_api.device_profile_create(
            context, obj_devprof)
        return DeviceProfile.convert_with_links(new_devprof)

    def _validate_post_request(self, req_devprof):
        NAME = "^[a-zA-Z0-9-_]+$"
        keys = '|'.join(["resources:", "trait:", "accel:"])
        GROUP_KEYS = "^(%s)" % keys
        TRAIT_VALUES = ["required", "forbidden"]

        name = req_devprof.get("name")
        if not name:
            raise exception.DeviceProfileNameNeeded()
        elif not re.match(NAME, name):
            raise exception.InvalidParameterValue(
                err="Device profile name must be of the form %s" % NAME)

        groups = req_devprof.get("groups")
        if not groups:
            raise exception.DeviceProfileGroupsExpected()

        for group in groups:
            tmp_group = copy.deepcopy(group)
            for key, value in tmp_group.items():
                # check resource and trait prefix format
                if not re.match(GROUP_KEYS, key):
                    raise exception.InvalidParameterValue(
                        err="Device profile group keys must be of"
                            " the form %s" % GROUP_KEYS)
                # check trait name and it's value
                if key.startswith("trait:"):
                    inner_origin_trait = ":".join(key.split(":")[1:])
                    inner_trait = inner_origin_trait.strip(" ")
                    if not inner_trait.startswith('CUSTOM_'):
                        raise exception.InvalidParameterValue(
                            err="Unsupported trait name format %s, should "
                                "start with CUSTOM_" % inner_trait)
                    if value not in TRAIT_VALUES:
                        raise exception.InvalidParameterValue(
                            err="Unsupported trait value %s, the value must"
                                " be one among %s" % (value, TRAIT_VALUES))
                    # strip " " and update old group key.
                    if inner_origin_trait != inner_trait:
                        del group[key]
                        standard_key = "trait:" + inner_trait
                        group[standard_key] = value
                # check rc name and it's value
                if key.startswith("resources:"):
                    inner_origin_rc = ":".join(key.split(":")[1:])
                    inner_rc = inner_origin_rc.strip(" ")
                    if inner_rc not in constants.SUPPORT_RESOURCES and \
                            not inner_rc.startswith('CUSTOM_'):
                        raise exception.InvalidParameterValue(
                            err="Unsupported resource class %s" % inner_rc)
                    try:
                        int(value)
                    except ValueError:
                        raise exception.InvalidParameterValue(
                            err="Resources number %s is invalid" % value)
                    # strip " " and update old group key.
                    if inner_origin_rc != inner_rc:
                        del group[key]
                        standard_key = "resources:" + inner_rc
                        group[standard_key] = value

    def _get_device_profile_list(self, names=None, uuid=None):
        """Get a list of API objects representing device profiles."""

        context = pecan.request.context
        obj_devprofs = objects.DeviceProfile.list(context)
        if names:
            new_obj_devprofs = [devprof for devprof in obj_devprofs
                                if devprof['name'] in names]
            obj_devprofs = new_obj_devprofs
        elif uuid is not None:
            new_obj_devprofs = [devprof for devprof in obj_devprofs
                                if devprof['uuid'] == uuid]
            obj_devprofs = new_obj_devprofs

        return obj_devprofs

    @authorize_wsgi.authorize_wsgi("cyborg:device_profile", "get_all", False)
    @expose.expose(DeviceProfileCollection, wtypes.text)
    def get_all(self, name=None):
        """Retrieve a list of device profiles."""
        if name is not None:
            names = name.split(',')
        else:
            names = []
        LOG.info('[device_profiles] get_all. names=%s', names)
        api_obj_devprofs = self._get_device_profile_list(names)

        ret = DeviceProfileCollection.convert_with_links(api_obj_devprofs)
        LOG.info('[device_profiles] get_all returned: %s', ret)
        return ret

    @authorize_wsgi.authorize_wsgi("cyborg:device_profile", "get_one")
    @expose.expose('json', wtypes.text)
    def get_one(self, dp_uuid_or_name):
        """Retrieve a single device profile by uuid or name."""
        context = pecan.request.context
        if uuidutils.is_uuid_like(dp_uuid_or_name):
            LOG.info('[device_profiles] get_one. uuid=%s', dp_uuid_or_name)
            obj_devprof = objects.DeviceProfile.get_by_uuid(context,
                                                            dp_uuid_or_name)
        else:
            if api.request.version.minor >= versions.MINOR_2_DP_BY_NAME:
                LOG.info('[device_profiles] get_one. name=%s', dp_uuid_or_name)
                obj_devprof = \
                    objects.DeviceProfile.get_by_name(context,
                                                      dp_uuid_or_name)
            else:
                raise exception.NotAcceptable(_(
                    "Request not acceptable. The minimal required API "
                    "version should be %(base)s.%(opr)s") %
                    {'base': versions.BASE_VERSION,
                     'opr': versions.MINOR_2_DP_BY_NAME})
        if not obj_devprof:
            LOG.warning("Device profile with %s not found!", dp_uuid_or_name)
            raise exception.ResourceNotFound(
                resource='Device profile',
                msg='with %s' % dp_uuid_or_name)
        api_obj_devprof = self.get_device_profile(obj_devprof)

        ret = {"device_profile": api_obj_devprof}
        LOG.info('[device_profiles] get_one returned: %s', ret)
        # TODO(Sundar) Replace this with convert_with_links()
        return wsme.api.Response(ret, status_code=HTTPStatus.OK,
                                 return_type=wsme.types.DictType)

    @authorize_wsgi.authorize_wsgi("cyborg:device_profile", "delete")
    @expose.expose(None, wtypes.text, status_code=HTTPStatus.NO_CONTENT)
    def delete(self, value):
        """Delete one or more device_profiles.

        URL: /device_profiles/{uuid} OR /device_profiles?value=foo,bar

        :param value: This should be one of these two:
            - UUID of a device_profile.
            - Comma-delimited list of device profile names.
        """
        context = pecan.request.context
        if uuidutils.is_uuid_like(value):
            uuid = value
            LOG.info('[device_profiles] delete uuid=%s', uuid)
            obj_devprof = objects.DeviceProfile.get_by_uuid(context, uuid)
            pecan.request.conductor_api.device_profile_delete(
                context, obj_devprof)
        else:
            names = value.split(",")
            LOG.info('[device_profiles] delete names=(%s)', names)
            for name in names:
                obj_devprof = objects.DeviceProfile.get_by_name(context, name)
                pecan.request.conductor_api.device_profile_delete(
                    context, obj_devprof)



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Version 2 of the Cyborg API"""

import pecan
from pecan import rest
from webob import exc
from wsme import types as wtypes

from cyborg.api import expose

from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers.v2 import arqs
from cyborg.api.controllers.v2 import attributes
from cyborg.api.controllers.v2 import deployables
from cyborg.api.controllers.v2 import device_profiles
from cyborg.api.controllers.v2 import devices

from cyborg.api.controllers.v2 import versions


def min_version():
    return base.Version(
        {base.Version.current_api_version: versions.min_version_string()},
        versions.min_version_string(), versions.max_version_string())


def max_version():
    return base.Version(
        {base.Version.current_api_version: versions.max_version_string()},
        versions.min_version_string(), versions.max_version_string())


class V2(base.APIBase):
    """The representation of the version 2 of the API."""

    id = wtypes.text
    """The ID of the version"""

    links = [link.Link]
    """Links to the accelerator resource"""

    max_version = wtypes.text
    """Highest microversion supported"""

    min_version = wtypes.text
    """Lowest microversion supported"""

    status = wtypes.text
    """Status"""

    @staticmethod
    def convert():
        v2 = V2()
        v2.id = 'v2.0'
        v2.max_version = str(max_version())
        v2.min_version = str(min_version())
        v2.status = 'CURRENT'
        v2.links = [
            link.Link.make_link('self', pecan.request.public_url,
                                '', ''),
            ]
        return v2


class Controller(rest.RestController):
    """Version 2 API controller root"""

    device_profiles = device_profiles.DeviceProfilesController()
    accelerator_requests = arqs.ARQsController()
    devices = devices.DevicesController()
    deployables = deployables.DeployablesController()
    attributes = attributes.AttributesController()

    @expose.expose(V2)
    def get(self):
        return V2.convert()

    def _check_version(self, version, headers=None):
        if headers is None:
            headers = {}
        # ensure that major version in the URL matches the header
        if version.major != versions.BASE_VERSION:
            raise exc.HTTPNotAcceptable(
                "Mutually exclusive versions requested. Version %(ver)s "
                "requested but not supported by this service. The supported "
                "version range is: [%(min)s, %(max)s]." %
                {'ver': version, 'min': versions.min_version_string(),
                 'max': versions.max_version_string()},
                headers=headers)
        # ensure the minor version is within the supported range
        if version < min_version() or version > max_version():
            raise exc.HTTPNotAcceptable(
                "Version %(ver)s was requested but the minor version is not "
                "supported by this service. The supported version range is: "
                "[%(min)s, %(max)s]." %
                {'ver': version, 'min': versions.min_version_string(),
                 'max': versions.max_version_string()},
                headers=headers)

    @pecan.expose()
    def _route(self, args, request=None):
        v = base.Version(pecan.request.headers, versions.min_version_string(),
                         versions.max_version_string())

        # The Vary header is used as a hint to caching proxies and user agents
        # that the response is also dependent on the OpenStack-API-Version and
        # not just the body and query parameters. See RFC 7231 for details.
        pecan.response.headers['Vary'] = base.Version.current_api_version

        # Always set the min and max headers
        pecan.response.headers[base.Version.min_api_version] = (
            versions.min_version_string())
        pecan.response.headers[base.Version.max_api_version] = (
            versions.max_version_string())

        # assert that requested version is supported
        self._check_version(v, pecan.response.headers)
        pecan.response.headers[base.Version.current_api_version] = str(v)
        pecan.request.version = v

        return super(Controller, self)._route(args, request)


__all__ = ('Controller',)



"""placeholder

Revision ID: 62bcf2610c5d
Revises: 7b696fd94949
Create Date: 2019-12-03 00:56:52.488257

"""

# revision identifiers, used by Alembic.
revision = '62bcf2610c5d'
down_revision = '7b696fd94949'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# Copyright 2022 Inspur, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import fixtures
from unittest import mock

from cyborg.common import exception
from cyborg.common import placement_client
from cyborg.tests import base


class PlacementAPITest(base.TestCase):

    def setUp(self):
        super(PlacementAPITest, self).setUp()
        self.instance_uuid = '00000000-0000-0000-0000-000000000001'

        self.mock_sdk = self.useFixture(fixtures.MockPatch(
            'cyborg.common.utils.get_sdk_adapter')).mock.return_value
        self.mock_log_info = self.useFixture(fixtures.MockPatch(
            'cyborg.common.placement_client.LOG.info')).mock
        self.mock_log_debug = self.useFixture(fixtures.MockPatch(
            'cyborg.common.placement_client.LOG.debug')).mock

    def test_get(self):
        self.mock_sdk.get.return_value = mock.Mock(status_code=200)
        placement = placement_client.PlacementClient()
        placement.get(mock.Mock())
        msg = 'Successfully get resources from placement: %s'
        self.mock_log_debug.assert_called_once_with(msg, mock.ANY)

    def test_get_exception(self):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        self.mock_sdk.get.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement.get, mock.Mock())

    def test_post(self):
        self.mock_sdk.post.return_value = mock.Mock(status_code=200)
        placement = placement_client.PlacementClient()
        placement.post(mock.Mock(), mock.ANY)
        msg = 'Successfully create resources from placement: %s'
        self.mock_log_debug.assert_called_once_with(msg, mock.ANY)

    def test_post_exception(self):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        self.mock_sdk.post.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement.post, mock.Mock(), mock.ANY)

    def test_put(self):
        self.mock_sdk.put.return_value = mock.Mock(status_code=200)
        placement = placement_client.PlacementClient()
        placement.put(mock.Mock(), mock.ANY)
        msg = 'Successfully update resources from placement: %s'
        self.mock_log_debug.assert_called_once_with(msg, mock.ANY)

    def test_put_exception(self):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        self.mock_sdk.put.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement.put, mock.Mock(), mock.ANY)

    def test_delete(self):
        self.mock_sdk.delete.return_value = mock.Mock(status_code=200)
        placement = placement_client.PlacementClient()
        placement.delete(mock.Mock(), mock.ANY)
        msg = 'Successfully delete resources from placement: %s'
        self.mock_log_debug.assert_called_once_with(msg, mock.ANY)

    def test_delete_exception(self):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        self.mock_sdk.delete.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement.delete, mock.Mock(), mock.ANY)

    def test_get_rp_traits(self):
        self.mock_sdk.get.return_value = mock.Mock(status_code=200)
        placement = placement_client.PlacementClient()
        placement._get_rp_traits(mock.ANY)
        msg = 'Successfully get resources from placement: %s'
        self.mock_log_debug.assert_called_once_with(msg, mock.ANY)

    def test_get_rp_traits_exception(self):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        self.mock_sdk.get.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement._get_rp_traits, mock.ANY)

    def test_ensure_traits(self):
        self.mock_sdk.put.return_value = mock.Mock(status_code=201)
        placement = placement_client.PlacementClient()
        self.mock_sdk.get.return_value = None
        placement._ensure_traits([mock.ANY])
        self.assertEqual(2, self.mock_log_debug.call_count)
        msg = 'Successfully update resources from placement: %s'
        # first call/arg
        self.assertIn(msg, self.mock_log_debug.call_args_list[1][0][0])

    def test_ensure_traits_exception(self):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        self.mock_sdk.get.return_value = None
        self.mock_sdk.put.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement._ensure_traits, [mock.ANY])

    @mock.patch('cyborg.common.placement_client.'
                'PlacementClient.get_resource_provider')
    def test_put_rp_traits(self, rp):
        self.mock_sdk.put.return_value = mock.Mock(status_code=200)
        placement = placement_client.PlacementClient()
        rp.return_value = {'status_code': 200, 'generation': 0}
        placement._put_rp_traits(mock.ANY, {'traits': 'fake_trait'})
        msg = 'Successfully update resources from placement: %s'
        self.mock_log_debug.assert_called_once_with(msg, mock.ANY)

    @mock.patch('cyborg.common.placement_client.'
                'PlacementClient.get_resource_provider')
    def test_put_rp_traits_exception(self, rp):
        placement = placement_client.PlacementClient()
        mock_ret = mock.Mock(status_code=500)
        rp.return_value = {'status_code': 200, 'generation': 0}
        self.mock_sdk.put.return_value = mock_ret
        self.assertRaises(exception.PlacementServerError,
                          placement._put_rp_traits,
                          mock.ANY, {'traits': 'fake_trait'})



# Copyright 2019 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_db import exception as db_exc

from cyborg.common import exception
from cyborg import objects
from cyborg.tests.unit.db.base import DbTestCase
from cyborg.tests.unit import fake_deployable
from cyborg.tests.unit import fake_device
from cyborg.tests.unit.objects import test_objects


class TestDeployableObject(DbTestCase):

    @property
    def fake_device(self):
        db_device = fake_device.get_fake_devices_as_dict()[0]
        return db_device

    @property
    def fake_deployable(self):
        db_deploy = fake_deployable.fake_db_deployable(id=1)
        return db_deploy

    @property
    def fake_deployable2(self):
        db_deploy = fake_deployable.fake_db_deployable(id=2)
        return db_deploy

    def test_create(self):
        db_device = self.fake_device
        device = objects.Device(context=self.context,
                                **db_device)
        device.create(self.context)
        device_get = objects.Device.get(self.context, device.uuid)
        db_dpl = self.fake_deployable
        dpl = objects.Deployable(context=self.context,
                                 **db_dpl)

        dpl.device_id = device_get.id
        dpl.create(self.context)

        self.assertEqual(db_dpl['uuid'], dpl.uuid)

    def test_get(self):
        db_device = self.fake_device
        device = objects.Device(context=self.context,
                                **db_device)
        device.create(self.context)
        device_get = objects.Device.get(self.context, device.uuid)
        db_dpl = self.fake_deployable
        dpl = objects.Deployable(context=self.context,
                                 **db_dpl)

        dpl.device_id = device_get.id
        dpl.create(self.context)
        dpl_get = objects.Deployable.get(self.context, dpl.uuid)
        self.assertEqual(dpl_get.uuid, dpl.uuid)

    def test_get_by_filter(self):
        db_device = self.fake_device
        device = objects.Device(context=self.context,
                                **db_device)
        device.create(self.context)
        device_get = objects.Device.get(self.context, device.uuid)
        db_dpl = self.fake_deployable
        dpl = objects.Deployable(context=self.context,
                                 **db_dpl)

        dpl.device_id = device_get.id
        dpl.create(self.context)
        query = {"uuid": dpl['uuid']}
        dpl_get_list = objects.Deployable.get_by_filter(self.context, query)

        self.assertEqual(dpl_get_list[0].uuid, dpl.uuid)

    def test_save(self):
        db_device = self.fake_device
        device = objects.Device(context=self.context,
                                **db_device)
        device.create(self.context)
        device_get = objects.Device.get(self.context, device.uuid)
        db_dpl = self.fake_deployable
        dpl = objects.Deployable(context=self.context,
                                 **db_dpl)

        dpl.device_id = device_get.id
        dpl.create(self.context)
        dpl.num_accelerators = 8
        dpl.save(self.context)
        dpl_get = objects.Deployable.get(self.context, dpl.uuid)
        self.assertEqual(dpl_get.num_accelerators, 8)

    def test_destroy(self):
        db_device = self.fake_device
        device = objects.Device(context=self.context,
                                **db_device)
        device.create(self.context)
        device_get = objects.Device.get(self.context, device.uuid)
        db_dpl = self.fake_deployable
        dpl = objects.Deployable(context=self.context,
                                 **db_dpl)

        dpl.device_id = device_get.id
        dpl.create(self.context)
        self.assertEqual(db_dpl['uuid'], dpl.uuid)
        dpl.destroy(self.context)
        self.assertRaises(exception.ResourceNotFound,
                          objects.Deployable.get, self.context,
                          dpl.uuid)


class TestDeployableObject(test_objects._LocalTest,
                           TestDeployableObject):
    def _test_save_objectfield_fk_constraint_fails(self, foreign_key,
                                                   expected_exception):

        error = db_exc.DBReferenceError('table', 'constraint', foreign_key,
                                        'key_table')
        # Prevent lazy-loading any fields, results in InstanceNotFound
        deployable = fake_deployable.fake_deployable_obj(self.context)
        fields_with_save_methods = [field for field in deployable.fields
                                    if hasattr(deployable, '_save_%s' % field)]
        for field in fields_with_save_methods:
            @mock.patch.object(deployable, '_save_%s' % field)
            @mock.patch.object(deployable, 'obj_attr_is_set')
            def _test(mock_is_set, mock_save_field):
                mock_is_set.return_value = True
                mock_save_field.side_effect = error
                deployable.obj_reset_changes(fields=[field])
                deployable._changed_fields.add(field)
                self.assertRaises(expected_exception, deployable.save)
                deployable.obj_reset_changes(fields=[field])
            _test()



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating Device via the DB API"""

import sys

from oslo_utils import uuidutils

from cyborg.common import exception
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils


class TestDbDevice(base.DbTestCase):

    def test_create(self):
        random_uuid = uuidutils.generate_uuid()
        kw = {'uuid': random_uuid}
        created_device = utils.create_test_device(self.context, **kw)
        self.assertEqual(random_uuid, created_device['uuid'])

    def test_get_by_uuid(self):
        created_dev = utils.create_test_device(self.context)
        queried_dev = self.dbapi.device_get(
            self.context, created_dev['uuid'])
        self.assertEqual(created_dev['uuid'], queried_dev['uuid'])

    def test_get_by_id(self):
        created_dev = utils.create_test_device(self.context)
        queried_dev = self.dbapi.device_get_by_id(
            self.context, created_dev['id'])
        self.assertEqual(created_dev['id'], queried_dev['id'])

    def test_update(self):
        created_dev = utils.create_test_device(self.context)
        queried_dev = self.dbapi.device_update(
            self.context, created_dev['uuid'], {'hostname': 'myhost'})
        self.assertEqual('myhost', queried_dev['hostname'])

    def test_list(self):
        uuids = []
        for i in range(1, 4):
            dev = utils.create_test_device(
                self.context,
                id=i,
                uuid=uuidutils.generate_uuid())
            uuids.append(dev['uuid'])
        devs = self.dbapi.device_list(self.context)
        dev_uuids = [item.uuid for item in devs]
        self.assertEqual(sorted(uuids), sorted(dev_uuids))

    def test_list_by_filters(self):
        dev1 = utils.create_test_device(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            hostname='myhost1')
        utils.create_test_device(
            self.context,
            id=2,
            uuid=uuidutils.generate_uuid(),
            hostname='myhost2')
        res = self.dbapi.device_list_by_filters(
            self.context, filters={"hostname": "myhost1"})
        self.assertEqual(1, len(res))
        self.assertEqual(dev1['hostname'], res[0]['hostname'])

    def test_delete(self):
        created_dev = utils.create_test_device(self.context)
        return_value = self.dbapi.device_delete(
            self.context,
            created_dev['uuid'])
        self.assertIsNone(return_value)

    def test_list_filter_is_none(self):
        """The main test is filters=None. If filters=None,
        it will be initialized to {}, that will return all device
        same as the List Device API response.
        """
        dev1 = utils.create_test_device(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid())
        res = self.dbapi.device_list_by_filters(
            self.context, filters=None)
        self.assertEqual(1, len(res))
        self.assertEqual(dev1['uuid'], res[0]['uuid'])

    def test_get_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_get,
                          self.context, random_uuid)

    def test_get_by_id_not_exist(self):
        fake_id = sys.maxsize
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_get_by_id,
                          self.context, fake_id)

    def test_delete_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_delete,
                          self.context, random_uuid)



# Copyright 2020 Inspur, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_serialization import jsonutils
from unittest import mock

from cyborg.accelerator.drivers.fpga.inspur.driver import InspurFPGADriver
from cyborg.accelerator.drivers.fpga.inspur import sysinfo
from cyborg.tests import base

INSPUR_FPGA_INFO = ("0000:86:00.0 Processing accelerators [1200]: "
                    "Inspur Electronic Information Industry Co., Ltd. "
                    "Device [1bd4:a115] (rev 04)")


class stdout(object):
    def readlines(self):
        return [INSPUR_FPGA_INFO]


class p(object):
    def __init__(self):
        self.stdout = stdout()

    def wait(self):
        pass


class TestInspurFPGADriver(base.TestCase):

    def setUp(self):
        super(TestInspurFPGADriver, self).setUp()
        self.p = p()

    @mock.patch('cyborg.accelerator.drivers.fpga.'
                'inspur.sysinfo.lspci_privileged')
    def test_discover(self, mock_devices_for_vendor):
        mock_devices_for_vendor.return_value = self.p.stdout.readlines()
        self.set_defaults(host='host-192-168-32-195', debug=True)
        fpga_list = InspurFPGADriver().discover()
        self.assertEqual(1, len(fpga_list))
        attach_handle_list = [
            {'attach_type': 'PCI',
             'attach_info': '{"bus": "86", '
                            '"device": "00", '
                            '"domain": "0000", '
                            '"function": "0"}',
             'in_use': False}
        ]
        attribute_list = [
            {'key': 'rc', 'value': 'FPGA'},
            {'key': 'trait0', 'value': 'CUSTOM_FPGA_INSPUR'},
            {'key': 'trait1', 'value': 'CUSTOM_FPGA_PRODUCT_ID_A115'},
        ]
        expected = {
            'vendor': '1bd4',
            'type': 'FPGA',
            'std_board_info': {"controller": "Processing accelerators",
                               "product_id": "a115"},
            'vendor_board_info': {"vendor_info": "fpga_vb_info"},
            'deployable_list':
                [
                    {
                        'num_accelerators': 1,
                        'driver_name': 'INSPUR',
                        'name': 'host-192-168-32-195_0000:86:00.0',
                        'attach_handle_list': attach_handle_list,
                        'attribute_list': attribute_list
                    },
                ],
            'controlpath_id': {'cpid_info': '{"bus": "86", '
                                            '"device": "00", '
                                            '"domain": "0000", '
                                            '"function": "0"}',
                               'cpid_type': 'PCI'}
        }
        fpga_obj = fpga_list[0]
        fpga_dict = fpga_obj.as_dict()
        fpga_dep_list = fpga_dict['deployable_list']
        fpga_attach_handle_list = (
            fpga_dep_list[0].as_dict()['attach_handle_list'])
        fpga_attribute_list = fpga_dep_list[0].as_dict()['attribute_list']
        attri_obj_data = []
        [attri_obj_data.append(attr.as_dict()) for attr in fpga_attribute_list]
        attribute_actual_data = sorted(attri_obj_data, key=lambda i: i['key'])
        self.assertEqual(expected['vendor'], fpga_dict['vendor'])
        self.assertEqual(expected['controlpath_id'],
                         fpga_dict['controlpath_id'])
        self.assertEqual(expected['std_board_info'],
                         jsonutils.loads(fpga_dict['std_board_info']))
        self.assertEqual(expected['vendor_board_info'],
                         jsonutils.loads(fpga_dict['vendor_board_info']))
        self.assertEqual(expected['deployable_list'][0]['num_accelerators'],
                         fpga_dep_list[0].as_dict()['num_accelerators'])
        self.assertEqual(expected['deployable_list'][0]['name'],
                         fpga_dep_list[0].as_dict()['name'])
        self.assertEqual(expected['deployable_list'][0]['driver_name'],
                         fpga_dep_list[0].as_dict()['driver_name'])
        self.assertEqual(attach_handle_list[0],
                         fpga_attach_handle_list[0].as_dict())
        self.assertEqual(attribute_list, attribute_actual_data)

    @mock.patch('cyborg.accelerator.drivers.fpga.'
                'inspur.sysinfo.lspci_privileged')
    def test_get_pci_devices_by_inspur_vendor(self, mock_devices_for_vendor):

        fake_pci_output = [("0000:86:00.0 Processing accelerators [1200]: "
                            "Inspur Electronic Information Industry Co., Ltd. "
                            "Device [1bd4:a115] (rev 04)\n"
                            "0000:86:00.0 Processing accelerators [1200]: "
                            "Xilinx Corporation Device [10ee:5000]")]
        mock_devices_for_vendor.return_value = fake_pci_output
        pci_devices = sysinfo.get_pci_devices(sysinfo.INSPUR_FPGA_FLAGS,
                                              vendor_id=sysinfo.VENDOR_ID)
        expected = [("0000:86:00.0 Processing accelerators [1200]: "
                     "Inspur Electronic Information Industry Co., Ltd. "
                     "Device [1bd4:a115] (rev 04)")]
        self.assertEqual(len(pci_devices), 1)
        self.assertEqual(pci_devices, expected)



# Copyright 2018 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import base as object_base

from cyborg.common import exception
from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields


LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class Attribute(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.UUIDField(nullable=False),
        'deployable_id': object_fields.IntegerField(nullable=False),
        'key': object_fields.StringField(nullable=False),
        'value': object_fields.StringField(nullable=False)
    }

    def create(self, context):
        """Create an attribute record in the DB."""
        if self.deployable_id is None:
            raise exception.AttributeInvalid()

        values = self.obj_get_changes()
        db_attr = self.dbapi.attribute_create(context,
                                              values)
        return self._from_db_object(self, db_attr)

    @classmethod
    def get(cls, context, uuid):
        """Find a DB attribute and return an Obj Deployable."""
        db_attr = cls.dbapi.attribute_get(context, uuid)
        obj_attr = cls._from_db_object(cls(context), db_attr)
        return obj_attr

    @classmethod
    def get_by_deployable_id(cls, context, deployable_id):
        """Get an attribute by deployable_id"""
        db_attr = cls.dbapi.attribute_get_by_deployable_id(context,
                                                           deployable_id)
        return cls._from_db_object_list(db_attr, context)

    @classmethod
    def get_by_filter(cls, context, filters):
        """Get an attribute by specified filters"""
        db_attr = cls.dbapi.attribute_get_by_filter(context, filters)
        return cls._from_db_object_list(db_attr, context)

    def save(self, context):
        """Update an attribute record in the DB."""
        db_attr = self.dbapi.attribute_update(context,
                                              self.uuid,
                                              self.key,
                                              self.value)
        self._from_db_object(self, db_attr)

    def destroy(self, context):
        """Delete an attribute from the DB."""
        self.dbapi.attribute_delete(context, self.uuid)
        self.obj_reset_changes()

    def set_key_value_pair(self, set_key, set_value):
        self.key = set_key
        self.value = set_value

    @classmethod
    def get_by_dep_key(cls, context, dep_id, key):
        """Get the only one attribute with deployable_id and the key."""
        attr_filter = {"deployable_id": dep_id, "key": key}
        attr_list = cls.get_by_filter(context, attr_filter)
        if len(attr_list) != 0:
            return attr_list[0]
        else:
            return None



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""initial migration.

Revision ID: f50980397351
Revises: None
Create Date: 2017-08-15 08:44:36.010417

"""

# revision identifiers, used by Alembic.
revision = 'f50980397351'
down_revision = None


from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'accelerators',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('project_id', sa.String(length=36), nullable=True),
        sa.Column('user_id', sa.String(length=36), nullable=True),
        sa.Column('device_type', sa.Text(), nullable=False),
        sa.Column('acc_type', sa.Text(), nullable=True),
        sa.Column('acc_capability', sa.Text(), nullable=True),
        sa.Column('vendor_id', sa.Text(), nullable=False),
        sa.Column('product_id', sa.Text(), nullable=False),
        sa.Column('remotable', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_accelerators0uuid'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'deployables',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('parent_uuid', sa.String(length=36),
                  sa.ForeignKey('deployables.uuid'), nullable=True),
        sa.Column('root_uuid', sa.String(length=36),
                  sa.ForeignKey('deployables.uuid'), nullable=True),
        sa.Column('address', sa.Text(), nullable=False),
        sa.Column('host', sa.Text(), nullable=False),
        sa.Column('board', sa.Text(), nullable=False),
        sa.Column('vendor', sa.Text(), nullable=False),
        sa.Column('version', sa.Text(), nullable=False),
        sa.Column('type', sa.Text(), nullable=False),
        sa.Column('interface_type', sa.Text(), nullable=False),
        sa.Column('assignable', sa.Boolean(), nullable=False),
        sa.Column('instance_uuid', sa.String(length=36), nullable=True),
        sa.Column('availability', sa.Text(), nullable=False),
        sa.Column('accelerator_id', sa.Integer(),
                  sa.ForeignKey('accelerators.id', ondelete="CASCADE"),
                  nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_deployables0uuid'),
        sa.Index('deployables_parent_uuid_idx', 'parent_uuid'),
        sa.Index('deployables_root_uuid_idx', 'root_uuid'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'attributes',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('deployable_id', sa.Integer(),
                  sa.ForeignKey('deployables.id', ondelete="CASCADE"),
                  nullable=False),
        sa.Column('key', sa.Text(), nullable=False),
        sa.Column('value', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_attributes0uuid'),
        sa.Index('attributes_deployable_id_idx', 'deployable_id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )



"""placeholder

Revision ID: 7a4fd0fc3f8c
Revises: 62bcf2610c5d
Create Date: 2019-12-03 00:57:01.579012

"""

# revision identifiers, used by Alembic.
revision = '7a4fd0fc3f8c'
down_revision = '62bcf2610c5d'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg GPU driver implementation.
"""
from oslo_log import log as logging

from cyborg.accelerator.drivers.gpu import utils


LOG = logging.getLogger(__name__)

VENDOR_MAPS = {"10de": "nvidia", "102b": "matrox"}


class GPUDriver(object):
    """Base class for GPU drivers.

       This is just a virtual GPU drivers interface.
       Vendor should implement their specific drivers.
    """

    @classmethod
    def create(cls, vendor, *args, **kwargs):
        for sclass in cls.__subclasses__():
            vendor_name = VENDOR_MAPS.get(vendor, vendor)
            if vendor_name == sclass.VENDOR:
                return sclass(*args, **kwargs)
        raise LookupError("Not find the GPU driver for vendor %s" % vendor)

    def discover(self):
        """Discover GPU information of current vendor(Identified by class).

        :return: List of GPU information dict.
        """
        raise NotImplementedError()

    @classmethod
    def discover_vendors(cls):
        """Discover GPU vendors of current node.

        :return: GPU vendor ID list.
        """
        return utils.discover_vendors()



# Copyright (c) 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


# This is the version 2 API
BASE_VERSION = 2


# Here goes a short log of changes in every version.
# Refer to cyborg/api/rest_api-version-history.rst for a detailed
# explanation of what each version contains.
#
# v2.0: Initial minor version.
# v2.1: Add project_id for arq patch
# v2.2: Support getting device profile by name (newly introduced) and uuid.
# v2.3: Add status info for device API.
MINOR_0_INITIAL_VERSION = 0
MINOR_1_PROJECT_ID = 1
MINOR_2_DP_BY_NAME = 2
MINOR_3_DEVICE_STATUS = 3

# When adding another version, update:
# - MINOR_MAX_VERSION
# - cyborg/api/rest_api-version-history.rst with a detailed
#   explanation of what changed in the new version


MINOR_MAX_VERSION = MINOR_3_DEVICE_STATUS

# String representations of the minor and maximum versions
_MIN_VERSION_STRING = '{}.{}'.format(BASE_VERSION, MINOR_0_INITIAL_VERSION)
_MAX_VERSION_STRING = '{}.{}'.format(BASE_VERSION, MINOR_MAX_VERSION)


def service_type_string():
    return 'accelerator'


def min_version_string():
    """Returns the minimum supported API version (as a string)"""
    return _MIN_VERSION_STRING


def max_version_string():
    """Returns the maximum supported API version (as a string).

    If the service is pinned, the maximum API version is the pinned
    version. Otherwise, it is the maximum supported API version.

    """
    return _MAX_VERSION_STRING



# -*- coding: utf-8 -*-

#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class NotificationEndpoint(object):
    # filter_rule = messaging.NotificationFilter(publisher_id='^cyborg.*')

    # We have an update from an agent and we need to add it to our in memory
    # cache of accelerator objects and schedule a flush to the database
    def update(self, ctxt, publisher_id, event_type, payload, metadata):
        print("Got update")
        return True

    # We have an info message from an agent, anything that wouldn't
    # go into the db but needs to be communicated goes here
    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        print("Got info")
        return True

    # We have a warning from an agent, we may take some action
    def warn(self, ctxt, publisher_id, event_type, payload, metadata):
        print("Got warn")
        return True

    # We have an error from an agent, we must take some action
    def error(self, ctxt, publisher_id, event_type, payload, metadata):
        print("Got error")
        return True



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating ExtArq via the DB API"""

from oslo_utils import uuidutils

from cyborg.common import exception
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils


class TestDbExtArq(base.DbTestCase):

    def test_create(self):
        random_uuid = uuidutils.generate_uuid()
        kw = {'uuid': random_uuid}
        created_extarq = utils.create_test_extarq(self.context, **kw)
        self.assertEqual(random_uuid, created_extarq['uuid'])

    def test_get_by_uuid(self):
        created_extarq = utils.create_test_extarq(self.context)
        queried_extarq = self.dbapi.extarq_get(
            self.context, created_extarq['uuid'])
        self.assertEqual(created_extarq['uuid'], queried_extarq['uuid'])

    def test_update(self):
        created_extarq = utils.create_test_extarq(self.context)
        queried_extarq = self.dbapi.extarq_update(
            self.context, created_extarq['uuid'], {'state': 'Initial'})
        self.assertEqual('Initial', queried_extarq['state'])

    def test_list(self):
        uuids = []
        for i in range(1, 4):
            extarq = utils.create_test_extarq(
                self.context,
                id=i,
                uuid=uuidutils.generate_uuid())
            uuids.append(extarq['uuid'])
        extarqs = self.dbapi.extarq_list(self.context)
        extarq_uuids = [item.uuid for item in extarqs]
        self.assertEqual(sorted(uuids), sorted(extarq_uuids))

    def test_delete(self):
        created_extarq = utils.create_test_extarq(self.context)
        return_value = self.dbapi.extarq_delete(
            self.context,
            created_extarq['uuid'])
        self.assertIsNone(return_value)

    def test_get_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.extarq_get,
                          self.context, random_uuid)

    def test_delete_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.extarq_delete,
                          self.context, random_uuid)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

from concurrent.futures import ThreadPoolExecutor as CFThreadPoolExecutor
from functools import wraps
import queue
import time
import traceback

from keystoneauth1 import exceptions as ks_exc
from keystoneauth1 import loading as ks_loading
from openstack import connection
from openstack import exceptions as sdk_exc
from os_service_types import service_types
from oslo_concurrency import lockutils
from oslo_log import log

from cyborg.common import exception
from cyborg.common.i18n import _
import cyborg.conf


LOG = log.getLogger(__name__)

synchronized = lockutils.synchronized_with_prefix('cyborg-')
_SERVICE_TYPES = service_types.ServiceTypes()
CONF = cyborg.conf.CONF


def safe_rstrip(value, chars=None):
    """Removes trailing characters from a string if that does not make it empty

    :param value: A string value that will be stripped.
    :param chars: Characters to remove.
    :return: Stripped value.

    """
    if not isinstance(value, str):
        LOG.warning("Failed to remove trailing character. Returning "
                    "original object. Supplied object is not a string: "
                    "%s,", value)
        return value

    return value.rstrip(chars) or value


def get_ksa_adapter(service_type, ksa_auth=None, ksa_session=None,
                    min_version=None, max_version=None):
    """Construct a keystoneauth1 Adapter for a given service type.

    We expect to find a conf group whose name corresponds to the service_type's
    project according to the service-types-authority.  That conf group must
    provide at least ksa adapter options.  Depending how the result is to be
    used, ksa auth and/or session options may also be required, or the relevant
    parameter supplied.

    :param service_type: String name of the service type for which the Adapter
                         is to be constructed.
    :param ksa_auth: A keystoneauth1 auth plugin. If not specified, we attempt
                     to find one in ksa_session.  Failing that, we attempt to
                     load one from the conf.
    :param ksa_session: A keystoneauth1 Session.  If not specified, we attempt
                        to load one from the conf.
    :param min_version: The minimum major version of the adapter's endpoint,
                        intended to be used as the lower bound of a range with
                        max_version.
                        If min_version is given with no max_version it is as
                        if max version is 'latest'.
    :param max_version: The maximum major version of the adapter's endpoint,
                        intended to be used as the upper bound of a range with
                        min_version.
    :return: A keystoneauth1 Adapter object for the specified service_type.
    :raise: ConfGroupForServiceTypeNotFound If no conf group name could be
            found for the specified service_type.
    """
    # Get the conf group corresponding to the service type.
    confgrp = _SERVICE_TYPES.get_project_name(service_type)
    if not confgrp or not hasattr(CONF, confgrp):
        # Try the service type as the conf group.  This is necessary for e.g.
        # placement, while it's still part of the nova project.
        # Note that this might become the first thing we try if/as we move to
        # using service types for conf group names in general.
        confgrp = service_type
        if not confgrp or not hasattr(CONF, confgrp):
            raise exception.ConfGroupForServiceTypeNotFound(stype=service_type)

    # Ensure we have an auth.
    # NOTE(efried): This could be None, and that could be okay - e.g. if the
    # result is being used for get_endpoint() and the conf only contains
    # endpoint_override.
    if not ksa_auth:
        if ksa_session and ksa_session.auth:
            ksa_auth = ksa_session.auth
        else:
            ksa_auth = ks_loading.load_auth_from_conf_options(CONF, confgrp)

    if not ksa_session:
        ksa_session = ks_loading.load_session_from_conf_options(
            CONF, confgrp, auth=ksa_auth)

    return ks_loading.load_adapter_from_conf_options(
        CONF, confgrp, session=ksa_session, auth=ksa_auth,
        min_version=min_version, max_version=max_version)


def _get_conf_group(service_type):
    # Get the conf group corresponding to the service type.
    confgrp = _SERVICE_TYPES.get_project_name(service_type)
    if not confgrp or not hasattr(CONF, confgrp):
        raise exception.ConfGroupForServiceTypeNotFound(stype=service_type)
    return confgrp


def _get_auth_and_session(confgrp):
    ksa_auth = ks_loading.load_auth_from_conf_options(CONF, confgrp)
    return ks_loading.load_session_from_conf_options(
        CONF, confgrp, auth=ksa_auth)


def get_sdk_adapter(service_type, check_service=False):
    """Construct an openstacksdk-brokered Adapter for a given service type.
    We expect to find a conf group whose name corresponds to the service_type's
    project according to the service-types-authority.  That conf group must
    provide ksa auth, session, and adapter options.
    :param service_type: String name of the service type for which the Adapter
                         is to be constructed.
    :param check_service: If True, we will query the endpoint to make sure the
            service is alive, raising ServiceUnavailable if it is not.
    :return: An openstack.proxy.Proxy object for the specified service_type.
    :raise: ConfGroupForServiceTypeNotFound If no conf group name could be
            found for the specified service_type.
    :raise: ServiceUnavailable if check_service is True and the service is down
    """
    confgrp = _get_conf_group(service_type)
    sess = _get_auth_and_session(confgrp)
    try:
        conn = connection.Connection(
            session=sess, oslo_conf=CONF, service_types={service_type},
            strict_proxies=check_service)
    except sdk_exc.ServiceDiscoveryException as e:
        raise exception.ServiceUnavailable(
            _("The %(service_type)s service is unavailable: %(error)s") %
            {'service_type': service_type, 'error': str(e)})
    return getattr(conn, service_type)


def get_endpoint(ksa_adapter):
    """Get the endpoint URL represented by a keystoneauth1 Adapter.

    This method is equivalent to what

        ksa_adapter.get_endpoint()

    should do, if it weren't for a panoply of bugs.

    :param ksa_adapter: keystoneauth1.adapter.Adapter, appropriately set up
                        with an endpoint_override; or service_type, interface
                        (list) and auth/service_catalog.
    :return: String endpoint URL.
    :raise EndpointNotFound: If endpoint discovery fails.
    """
    # TODO(efried): The remainder of this method reduces to
    # TODO(efried):     return ksa_adapter.get_endpoint()
    # TODO(efried): once bug #1709118 is fixed.
    # NOTE(efried): Id9bd19cca68206fc64d23b0eaa95aa3e5b01b676 may also do the
    #               trick, once it's in a ksa release.
    # The EndpointNotFound exception happens when _ContextAuthPlugin is in play
    # because its get_endpoint() method isn't yet set up to handle interface as
    # a list.  (It could also happen with a real auth if the endpoint isn't
    # there; but that's covered below.)
    try:
        return ksa_adapter.get_endpoint()
    except ks_exc.EndpointNotFound:
        pass

    interfaces = list(ksa_adapter.interface)
    for interface in interfaces:
        ksa_adapter.interface = interface
        try:
            return ksa_adapter.get_endpoint()
        except ks_exc.EndpointNotFound:
            pass
    raise ks_exc.EndpointNotFound(
        "Could not find requested endpoint for any of the following "
        "interfaces: %s" % interfaces)


class _Singleton(type):
    """A metaclass that creates a Singleton base class when called."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        ins = cls._instances.get(cls)
        if not ins or (hasattr(ins, "_reset")
                       and isinstance(ins, cls) and ins._reset()):
            cls._instances[cls] = super(
                _Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]


class Singleton(_Singleton('SingletonMeta', (object,), {})):
    """A class for Singleton pattern."""

    pass


class ThreadPoolExecutor(CFThreadPoolExecutor):
    """Derived from concurrent.futures.ThreadPoolExecutor"""

    def __init__(self, *args, **kwargs):
        """Initializes a new ThreadPoolExecutor instance.

        Args:
            max_workers: The maximum number of threads that can be used to
                execute the given calls.
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: A callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.
        """
        super(ThreadPoolExecutor, self).__init__(*args, **kwargs)
        # NOTE(Shaohe): py37/38 will use SimpleQueue as _work_queue, it will
        # cause hang the main thread with eventlet.monkey_patch. Change it
        # to queue._PySimpleQueue
        if hasattr(queue, "SimpleQueue") and not isinstance(
                self._work_queue, queue._PySimpleQueue):
            self._work_queue = queue._PySimpleQueue()


class ThreadWorks(Singleton):
    """Passthrough method for ThreadPoolExecutor.

    It will also grab the context from the threadlocal store and add it to
    the store on the new thread.  This allows for continuity in logging the
    context when using this method to spawn a new thread.
    """

    def __init__(self, pool_size=CONF.thread_pool_size):
        """Singleton ThreadWorks init."""
        # Ref: https://pythonhosted.org/futures/
        # NOTE(Shaohe) We can let eventlet greening ThreadPoolExecutor
        # eventlet.patcher.monkey_patch(os=False, socket=True,
        #     select=True, thread=True)
        # futures = eventlet.import_patched('concurrent.futures')
        # ThreadPoolExecutor = futures.ThreadPoolExecutor
        self.executor = ThreadPoolExecutor(max_workers=pool_size)
        self.masters = {}

    def spawn(self, func, *args, **kwargs):
        """Put a job in thread pool."""
        LOG.debug("Add an async jobs. func: %s is with parameters args: %s, "
                  "kwargs: %s", func, args, kwargs)
        future = self.executor.submit(func, *args, **kwargs)
        return future

    def spawn_master(self, func, *args, **kwargs):
        """Start a new thread for a job."""
        executor = ThreadPoolExecutor()
        # TODO(Shaohe) every submit func should be wrapped with exception catch
        job = executor.submit(func, *args, **kwargs)
        LOG.debug("Spawn master job. func: %s is with parameters args: %s, "
                  "kwargs: %s", func, args, kwargs)
        # NOTE(Shaohe) shutdown should be after job submit
        executor.shutdown(wait=False)
        # TODO(Shaohe) we need to consider resouce collection such as the
        # follow code to recoder them with timestemp?
        # master = {tag: {
        #     "executor": executor,
        #     "job": f,
        #     "timestemp": time.time(),
        #     "timeout": timeout}}
        # self.masters.update(master)
        return job

    def _reset(self):
        return self.executor._shutdown

    def map(self, func, *iterables, **kwargs):
        """Batch for job function."""
        return self.executor.map(func, *iterables, **kwargs)

    @classmethod
    def get_workers_result(cls, fs=(), **kwargs):
        """get a jobs worker result.

        Waits workers util it finish or raise any Exception.
        It will cancel the rest if one job worker fails.
        If the future is cancelled before completing then CancelledError
        will be raised.

        Parameters:
            fs: the workers list spawn return.
            timeout: Wait workers timeout, it can be an int or float.
                     If the worker hasn't yet completed then this method
                     will wait up to timeout seconds. If the worker hasn't
                     completed in timeout seconds, then a
                     concurrent.futures.TimeoutError will be raised.
                     If timeout is not specified or None, there is no limit
                     to the wait time.
        return a generator which include:
            result: the value returned by the job workers.
            exception: the exception details raised from workers.
            state: The work state.
        """

        def future_iterator():
            # Yield must be hidden in closure so that the futures are submitted
            # before the first iterator value is required.
            try:
                # reverse to keep finishing order
                fs.reverse()
                while fs:
                    # Careful not to keep a reference to the popped future
                    f = None
                    if timeout is None:
                        f = fs.pop()
                        yield f.result(), f.exception(), f._state, None
                    else:
                        f = fs.pop()
                        yield (f.result(end_time - time.time()),
                               f.exception(), f._state, None)
            except Exception as e:
                err = traceback.format_exc()
                LOG.error("Error during check the worker status. Exception "
                          "info: %s", err)
                if f:
                    LOG.error("Error during check the worker status. "
                              "Exception info: %s, result: %s, state: %s. "
                              "Reason %s", f.exception(), f._result,
                              f._state, str(e))
                    yield f._result, f.exception(), f._state, err
            finally:
                # Do best to cancel remain jobs.
                if fs:
                    LOG.info("Cancel the remained pending jobs")
                for future in fs:
                    future.cancel()

        timeout = kwargs.get('timeout')
        if timeout is not None:
            end_time = timeout + time.time()
            LOG.info("Job timeout set as %s", timeout)
        fs = list(fs)

        return future_iterator()


# info https://www.oreilly.com/library/view/python-cookbook/
# 0596001673/ch14s05.html
def format_tb(tb, limit=None):
    """Fromat traceback to a string list.

    Print the usual traceback information, followed by a listing of all the
    local variables in each frame.
    """
    if not tb:
        return []
    tbs = ['Traceback (most recent call last):\n']
    while 1:
        tbs = tbs + traceback.format_tb(tb, limit)
        if not tb.tb_next:
            break
        tb = tb.tb_next
    return tbs


def wrap_job_tb(msg="Reason: %s"):
    """Wrap a function with a is_job tag added, and catch Excetpion."""
    def _wrap_job_tb(method):
        @wraps(method)
        def _impl(self, *args, **kwargs):
            try:
                output = method(self, *args, **kwargs)
            except Exception as e:
                LOG.error(msg, str(e))
                LOG.error(traceback.format_exc())
                raise
            return output
        setattr(_impl, "is_job", True)
        return _impl
    return _wrap_job_tb


def factory_register(SuperClass, ClassName):
    """Register an concrete class to a factory Class."""
    def decorator(Class):
        # return Class
        if not hasattr(SuperClass, "_factory"):
            setattr(SuperClass, "_factory", {})
        SuperClass._factory[ClassName] = Class
        setattr(Class, "_factory_type", ClassName)
        return Class
    return decorator


class FactoryMixin(object):
    """A factory Mixin to create an concrete class."""

    @classmethod
    def factory(cls, typ, *args, **kwargs):
        """factory to create an concrete class."""
        f = getattr(cls, "_factory", {})
        sclass = f.get(typ, None)
        if sclass:
            LOG.info("Find %s of concrete %s by %s.",
                     sclass.__name__, cls.__name__, typ)
            return sclass
        for sclass in cls.__subclasses__():
            if typ == getattr(cls, "_factory_type", None):
                return sclass
        else:
            return cls
            LOG.info("Use default %s, do not find concrete class"
                     "by %s.", cls.__name__, typ)


def strtime(at):
    return at.strftime("%Y-%m-%dT%H:%M:%S.%f")



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os_resource_classes as orc

CONDUCTOR_TOPIC = 'cyborg-conductor'
AGENT_TOPIC = 'cyborg-agent'
DEVICE_GPU = 'GPU'
DEVICE_FPGA = 'FPGA'
DEVICE_AICHIP = 'AICHIP'
DEVICE_QAT = 'QAT'
DEVICE_NIC = 'NIC'
DEVICE_SSD = 'SSD'


ARQ_STATES = (ARQ_INITIAL, ARQ_BIND_STARTED, ARQ_BOUND, ARQ_UNBOUND,
              ARQ_BIND_FAILED, ARQ_UNBIND_FAILED, ARQ_DELETING) = (
    'Initial', 'BindStarted', 'Bound', 'Unbound', 'BindFailed', 'UnbindFailed',
    'Deleting')


ARQ_BIND_STAGE = (ARQ_PRE_BIND, ARQ_FINISH_BIND,
                  ARQ_OUFOF_BIND_FLOW) = (
    [ARQ_INITIAL, ARQ_BIND_STARTED],
    [ARQ_BOUND, ARQ_BIND_FAILED],
    [ARQ_UNBOUND, ARQ_DELETING])


ARQ_BIND_STATUS = (ARQ_BIND_STATUS_FINISH, ARQ_BIND_STATUS_FAILED) = (
    "completed", "failed")


ARQ_BIND_STATES_STATUS_MAP = {
    ARQ_BOUND: ARQ_BIND_STATUS_FINISH,
    ARQ_BIND_FAILED: ARQ_BIND_STATUS_FAILED,
    ARQ_DELETING: ARQ_BIND_STATUS_FAILED
}

# TODO(Shaohe): maybe we can use oslo automaton lib
# ref: https://docs.openstack.org/automaton/latest/user/examples.html
# The states in value list can transfrom to the key state
ARQ_STATES_TRANSFORM_MATRIX = {
    ARQ_INITIAL: [],
    ARQ_BIND_STARTED: [ARQ_INITIAL, ARQ_UNBOUND],
    ARQ_BOUND: [ARQ_BIND_STARTED],
    ARQ_UNBOUND: [ARQ_INITIAL, ARQ_BIND_STARTED, ARQ_BOUND, ARQ_BIND_FAILED],
    ARQ_BIND_FAILED: [ARQ_BIND_STARTED, ARQ_BOUND],
    ARQ_DELETING: [ARQ_INITIAL, ARQ_BIND_STARTED, ARQ_BOUND,
                   ARQ_UNBOUND, ARQ_BIND_FAILED]
}


# Device type
DEVICE_TYPE = (DEVICE_GPU, DEVICE_FPGA, DEVICE_AICHIP, DEVICE_QAT, DEVICE_NIC,
               DEVICE_SSD)


# Device type
DEVICE_STATUS = ("enabled", "maintaining")


# Attach handle type
#  'TEST_PCI': used by fake driver, ignored by Nova virt driver.
ATTACH_HANDLE_TYPES = (AH_TYPE_PCI, AH_TYPE_MDEV, AH_TYPE_TEST_PCI) = (
    "PCI", "MDEV", "TEST_PCI")


# Control Path ID type
CPID_TYPE = (CPID_TYPE_PCI) = ("PCI")


# Resource Class
RESOURCES = {
    "FPGA": orc.FPGA,
    "PGPU": orc.PGPU,
    "VGPU": orc.VGPU,
    "QAT": "CUSTOM_QAT",
    "NIC": "CUSTOM_NIC",
    "SSD": 'CUSTOM_SSD',
}


ACCEL_SPECS = (
    ACCEL_BITSTREAM_ID,
    ACCEL_FUNCTION_ID
) = (
    "accel:bitstream_id",
    "accel:function_id"
)


SUPPORT_RESOURCES = (
    FPGA, GPU, VGPU, PGPU, QAT, NIC, SSD) = (
    "FPGA", "GPU", "VGPU", "PGPU", "CUSTOM_QAT", "CUSTOM_NIC", "CUSTOM_SSD"
)


FPGA_TRAITS = (
    FPGA_FUNCTION_ID,
) = (
    "CUSTOM_FPGA_FUNCTION_ID",
)


RESOURCES_PREFIX = "resources:"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock

from cyborg.accelerator.drivers.nic.intel import sysinfo

from cyborg.accelerator.drivers.nic.intel.driver import IntelNICDriver
from cyborg.tests import base
from cyborg.tests.unit.accelerator.drivers.nic.intel import prepare_test_data

import fixtures


class TestIntelNICDriver(base.TestCase):
    def setUp(self):
        super(TestIntelNICDriver, self).setUp()
        self.pcipath = sysinfo.PCI_DEVICES_PATH_PATTERN
        tmp_sys_dir = self.useFixture(fixtures.TempDir())
        prepare_test_data.create_fake_sysfs(tmp_sys_dir.path)
        tmp_path = tmp_sys_dir.path
        sysinfo.PCI_DEVICES_PATH_PATTERN = os.path.join(
            tmp_path, sysinfo.PCI_DEVICES_PATH_PATTERN.split("/", 1)[-1])

    def tearDown(self):
        super(TestIntelNICDriver, self).tearDown()
        sysinfo.PCI_DEVICES_PATH = self.pcipath

    @mock.patch("cyborg.accelerator.common.utils.get_ifname_by_pci_address")
    def test_discover(self, mock_device_ifname):
        mock_device_ifname.return_value = "ethx"
        attach_handle_list = [
            [
                {'attach_type': 'PCI',
                 'attach_info': '{"bus": "05", '
                                '"device": "01", '
                                '"domain": "0000", '
                                '"function": "0"}',
                 'in_use': False}
            ],
            [
                {'attach_type': 'PCI',
                 'attach_info': '{"bus": "06", '
                                '"device": "00", '
                                '"domain": "0000", '
                                '"function": "0"}',
                 'in_use': False}
            ]
        ]
        attribute_list = [
            [
                {
                    "key": "rc",
                    "value": "CUSTOM_NIC"
                    },
                {
                    "key": "trait0",
                    "value": "CUSTOM_VF"
                    }
            ],
            [
                {
                    "key": "rc",
                    "value": "CUSTOM_NIC"
                    },
                {
                    "key": "trait0",
                    "value": "CUSTOM_PF"
                    }
            ],
        ]
        expected = [{'vendor': '0x8086',
                     'type': 'NIC',
                     'deployable_list':
                         [
                             {'num_accelerators': 1,
                              'name': '0000:05:01.0',
                              'attach_handle_list': attach_handle_list[0],
                              'attribute_list':attribute_list[0]
                              },
                         ],
                     'controlpath_id':
                         {
                             'cpid_info': '{"bus": "05", '
                                          '"device": "00", '
                                          '"domain": "0000", '
                                          '"function": "0"}',
                             'cpid_type': 'PCI'}
                     },
                    {'vendor': '0x8086',
                     'type': 'NIC',
                     'deployable_list':
                         [
                             {'num_accelerators': 1,
                              'name': '0000:06:00.0',
                              'attach_handle_list': attach_handle_list[1],
                              'attribute_list':attribute_list[1]
                              },
                         ],
                     'controlpath_id':
                         {
                             'cpid_info': '{"bus": "06", '
                                          '"device": "00", '
                                          '"domain": "0000", '
                                          '"function": "0"}',
                             'cpid_type': 'PCI'}
                     }
                    ]
        intel = IntelNICDriver()
        nics = intel.discover()
        list.sort(nics, key=lambda x: x._obj_deployable_list[0].name)
        self.assertEqual(2, len(nics))
        for i in range(len(nics)):
            nic_dict = nics[i].as_dict()
            nic_dep_list = nic_dict['deployable_list']
            nic_attach_handle_list = \
                nic_dep_list[0].as_dict()['attach_handle_list']
            nic_attribute_list = \
                nic_dep_list[0].as_dict()['attribute_list']
            self.assertEqual(expected[i]['vendor'], nic_dict['vendor'])
            self.assertEqual(expected[i]['controlpath_id'],
                             nic_dict['controlpath_id'])
            self.assertEqual(expected[i]['deployable_list'][0]
                             ['num_accelerators'],
                             nic_dep_list[0].as_dict()['num_accelerators'])
            self.assertEqual(1, len(nic_attach_handle_list))
            self.assertEqual(attach_handle_list[i][0],
                             nic_attach_handle_list[0].as_dict())
            self.assertEqual(attribute_list[i][0],
                             nic_attribute_list[0].as_dict())



"""placeholder

Revision ID: 7b696fd94949
Revises: 22fb1af2d51e
Create Date: 2019-12-03 00:56:34.006320

"""

# revision identifiers, used by Alembic.
revision = '7b696fd94949'
down_revision = '22fb1af2d51e'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cyborg db test utilities."""

from cyborg.db import api as db_api


def get_test_deployable(**kw):
    return {
        'id': kw.get('id', 1),
        'uuid': kw.get('uuid', '10efe63d-dfea-4a37-ad94-4116fba50112'),
        'parent_id': kw.get('parent_id', None),
        'root_id': kw.get('root_id', 0),
        'name': kw.get('name', 'name'),
        'num_accelerators': kw.get('num_accelerators', 4),
        'device_id': kw.get('device_id', 0),
        'driver_name': kw.get('driver_name', 'NVIDIA'),
        'rp_uuid': kw.get('rp_uuid', '1c559644-2b56-3470-8427-d9d71f0f8621'),
        'bitstream_id': kw.get('bitstream_id', None),
        'created_at': kw.get('created_at', None),
        'updated_at': kw.get('updated_at', None)
    }


def create_test_deployable(context, **kwargs):
    """Create test deployable entry in DB and return Deployable DB object.

    Function to be used to create test Deployable objects in the database.
    :param context: request context.
    :param kwargs: kwargs which override default deployable attributes.
    :returns: Test Deployable DB object.
    """
    deployable = get_test_deployable(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.deployable_create(context, deployable)


def get_test_device(**kw):
    std_board_info = {
        "class": "Fake class",
        "device_id": "0xabcd"
        }
    return {
        'id': kw.get('id', 1),
        'uuid': kw.get('uuid', '20efe63d-dfea-4a37-ad94-4116fba50122'),
        'type': kw.get('type', "FPGA"),
        'vendor': kw.get('vendor', "0xABCD"),
        'model': kw.get('model', 'miss model info'),
        'std_board_info': kw.get('std_board_info', str(std_board_info)),
        'vendor_board_info': kw.get('vendor_board_info', 'fake_vendor_info'),
        'hostname': kw.get('hostname', "localhost"),
        'created_at': kw.get('created_at', None),
        'updated_at': kw.get('updated_at', None),
    }


def create_test_device(context, **kwargs):
    """Create test device entry in DB and return Device DB object.

    Function to be used to create test Device objects in the database.
    :param context: request context.
    :param kwargs: kwargs which override default device attributes.
    :returns: Test Device DB object.
    """
    device = get_test_device(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.device_create(context, device)


def get_test_extarq(**kwargs):
    return {
        'uuid': kwargs.get('uuid', '10efe63d-dfea-4a37-ad94-4116fba50986'),
        'id': kwargs.get('id', 1),
        'state': kwargs.get('state', 'Bound'),
        'device_profile_id': kwargs.get('id', 1),
        'hostname': kwargs.get('hostname', 'testnode1'),
        'device_rp_uuid': kwargs.get('device_rp_uuid',
                                     'f2b96c5f-242a-41a0-a736-b6e1fada071b'),
        'device_instance_uuid':
            kwargs.get('device_rp_uuid',
                       '6219e0fb-2935-4db2-a3c7-86a2ac3ac84e'),
        'attach_handle_id': kwargs.get('id', 1),
        'created_at': kwargs.get('created_at', None),
        'updated_at': kwargs.get('updated_at', None)
    }


def create_test_extarq(context, **kwargs):
    """Create test extend arq entry in DB and return ExtArq DB object.

    Function to be used to create test ExtArq objects in the database.
    :param context: request context.
    :param kwargs: kwargs which override default extend arq attributes.
    :returns: Test ExtArq DB object.
    """
    ext_arq = get_test_extarq(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.extarq_create(context, ext_arq)


def get_test_arq(**kwargs):
    return {
        'uuid': kwargs.get('uuid', '10efe63d-dfea-4a37-ad94-4116fba50986'),
        'id': kwargs.get('id', 1),
        'state': kwargs.get('state', 'Initial'),
        'device_profile': kwargs.get('device_profile', None),
        'hostname': kwargs.get('hostname', 'testnode1'),
        'device_rp_uuid': kwargs.get('device_rp_uuid',
                                     'f2b96c5f-242a-41a0-a736-b6e1fada071b'),
        'device_instance_uuid':
            kwargs.get('device_rp_uuid',
                       '6219e0fb-2935-4db2-a3c7-86a2ac3ac84e'),
        'attach_handle': kwargs.get('attach_handle', None),
        'created_at': kwargs.get('created_at', None),
        'updated_at': kwargs.get('updated_at', None),
        'substate': kwargs.get('substate', 'Initial'),
    }


def get_test_attach_handle(**kw):
    return {
        'uuid': kw.get('uuid', '10efe63d-dfea-4a37-ad94-4116fba50986'),
        'id': kw.get('id', 1),
        'deployable_id': kw.get('deployable_id', 1),
        'cpid_id': kw.get('cpid_id', 1),
        'in_use': kw.get('in_use', False),
        'attach_type': kw.get('attach_type', "PCI"),
        'attach_info': kw.get('attach_info', "attach_info"),
        'created_at': kw.get('create_at', None),
        'updated_at': kw.get('updated_at', None),
    }


def create_test_attach_handle(context, **kwargs):
    """Create test attach handle entry in DB and return
    AttachHandle DB object.

    Function to be used to create test AttachHandle objects in the database.
    :param context: request context.
    :param kwargs: kwargs which override default attach handle attributes.
    :returns: Test AttachHandle DB object.
    """
    attach_handle = get_test_attach_handle(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.attach_handle_create(context, attach_handle)


def get_test_control_path(**kw):
    return {
        'uuid': kw.get('uuid', '10efe63d-dfea-4a37-ad94-4116fba50986'),
        'id': kw.get('id', 1),
        'device_id': kw.get('device_id', 1),
        'cpid_type': kw.get('cpid_type', "PCI"),
        'cpid_info': kw.get('cpid_info',
                            '{"device": "2", "bus": "00", "function": "01", '
                            '"domain": "0001"}'),
        'created_at': kw.get('create_at', None),
        'updated_at': kw.get('updated_at', None),
    }


def get_test_device_profile(**kw):
    return {
        'id': kw.get('id', 1),
        'uuid': kw.get('uuid', 'c0f43d55-03bf-4831-8639-9bbdb6be2478'),
        'name': kw.get('name', 'name'),
        'description': kw.get('description', 'fake_dp_desc'),
        'profile_json': kw.get(
            'profile_json',
            '{"version": "1.0", \
             "groups": [{"resources:CUSTOM_ACCELERATOR_FPGA": "1"}, \
             {"trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10": "required"}, \
             {"trait:CUSTOM_FUNCTION_ID_3AFB": "required"}]}'),
        'created_at': kw.get('create_at', None),
        'updated_at': kw.get('updated_at', None),
    }


def create_test_device_profile(context, **kwargs):
    """Create test device profile entry in DB and return
    DeviceProfile DB object.

    Function to be used to create test DeviceProfile objects in the database.
    :param context: request context.
    :param kwargs: kwargs which override default device profile attributes.
    :returns: Test DeviceProfile DB object.
    """
    device_profile = get_test_device_profile(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.device_profile_create(context, device_profile)



"""add-device-status

Revision ID: 6c77bd6afea5
Revises: 4cc1d79978fc
Create Date: 2023-08-15 23:05:31.918963

"""

# revision identifiers, used by Alembic.
revision = '6c77bd6afea5'
down_revision = '4cc1d79978fc'

from alembic import op
import sqlalchemy as sa



def upgrade():
    new_column = sa.Column('status', sa.Enum('enabled', 'maintaining'),
                           nullable=False, default='enabled')
    op.add_column('devices', new_column)



"""placeholder

Revision ID: 57539722e5cf
Revises: c1b5abada09c
Create Date: 2019-12-03 00:55:16.012247

"""

# revision identifiers, used by Alembic.
revision = '57539722e5cf'
down_revision = 'c1b5abada09c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes for storage engines."""

import abc

from oslo_config import cfg
from oslo_db import api as db_api


_BACKEND_MAPPING = {'sqlalchemy': 'cyborg.db.sqlalchemy.api'}
IMPL = db_api.DBAPI.from_config(cfg.CONF,
                                backend_mapping=_BACKEND_MAPPING,
                                lazy=True)


def get_instance():
    """Return a DB API instance."""
    return IMPL


class Connection(object, metaclass=abc.ABCMeta):
    """Base class for storage system connections."""

    @abc.abstractmethod
    def __init__(self):
        """Constructor."""

    # device
    @abc.abstractmethod
    def device_create(self, context, values):
        """Create a new device when device is inserted into the host."""

    @abc.abstractmethod
    def device_get(self, context, uuid):
        """Get requested device."""

    @abc.abstractmethod
    def device_list(self, context, limit=None, marker=None,
                    sort_key=None, sort_dir=None):
        """Get requested list of devices."""

    @abc.abstractmethod
    def device_list_by_filters(self, context,
                               filters, sort_key='created_at',
                               sort_dir='desc', limit=None,
                               marker=None, columns_to_join=None):
        """Get requested devices by filters."""

    @abc.abstractmethod
    def device_update(self, context, uuid, values):
        """Update a device."""

    @abc.abstractmethod
    def device_delete(self, context, uuid):
        """Delete a device when device is removed from the host."""

    # device_profile
    @abc.abstractmethod
    def device_profile_create(self, context, values):
        """Create a new device_profile."""

    @abc.abstractmethod
    def device_profile_get_by_uuid(self, context, uuid):
        """Get requested device_profile by uuid."""

    @abc.abstractmethod
    def device_profile_get_by_id(self, context, id):
        """Get requested device_profile by id."""

    @abc.abstractmethod
    def device_profile_get(self, context, name):
        """Get requested device_profile by name."""

    @abc.abstractmethod
    def device_profile_list(self, context):
        """Get requested list of device_profiles."""

    @abc.abstractmethod
    def device_profile_list_by_filters(self, context,
                                       filters, sort_key='created_at',
                                       sort_dir='desc', limit=None,
                                       marker=None, columns_to_join=None):
        """Get requested list of device_profiles by filters."""

    @abc.abstractmethod
    def device_profile_update(self, context, uuid, values):
        """Update a device_profile."""

    @abc.abstractmethod
    def device_profile_delete(self, context, uuid):
        """Delete a device_profile."""

    # deployable
    @abc.abstractmethod
    def deployable_create(self, context, values):
        """Create a new deployable."""

    @abc.abstractmethod
    def deployable_get(self, context, uuid):
        """Get requested deployable."""

    @abc.abstractmethod
    def deployable_list(self, context):
        """Get requested list of deployables."""

    @abc.abstractmethod
    def deployable_update(self, context, uuid, values):
        """Update a deployable."""

    @abc.abstractmethod
    def deployable_delete(self, context, uuid):
        """Delete a deployable."""

    @abc.abstractmethod
    def deployable_get_by_filters(self, context,
                                  filters, sort_key='created_at',
                                  sort_dir='desc', limit=None,
                                  marker=None, columns_to_join=None):
        """Get requested deployable by filters."""

    @abc.abstractmethod
    def deployable_get_by_rp_uuid(self, context, rp_uuid):
        """Get requested deployable by resource provider UUID."""

    # attributes
    @abc.abstractmethod
    def attribute_create(self, context, values):
        """Create a new attribute."""

    @abc.abstractmethod
    def attribute_get(self, context, uuid):
        """Get requested attribute."""

    @abc.abstractmethod
    def attribute_get_by_deployable_id(self, context, deployable_id):
        """Get requested attribute by attribute id."""

    @abc.abstractmethod
    def attribute_get_by_filter(self, context, filters):
        """Get requested attribute by kv pair and attribute id."""

    @abc.abstractmethod
    def attribute_update(self, context, uuid, key, value):
        """Update an attribute's key value pair."""

    @abc.abstractmethod
    def attribute_delete(self, context, uuid):
        """Delete an attribute."""

    # quota
    @abc.abstractmethod
    def quota_reserve(self, context, resources, deltas, expire,
                      until_refresh, max_age, project_id=None,
                      is_allocated_reserve=False):
        """Check quotas and create appropriate reservations."""

    @abc.abstractmethod
    def reservation_commit(self, context, reservations, project_id=None):
        """Check quotas and create appropriate reservations."""

    # extarq
    @abc.abstractmethod
    def extarq_create(self, context, values):
        """Create a new extarq."""

    @abc.abstractmethod
    def extarq_delete(self, context, uuid):
        """Delete an extarq."""

    @abc.abstractmethod
    def extarq_update(self, context, uuid, values, state_scope=None):
        """Update an extarq."""

    @abc.abstractmethod
    def extarq_list(self, context, uuid_range=None):
        """Get requested list of extarqs."""

    @abc.abstractmethod
    def extarq_get(self, context, uuid, lock=False):
        """Get requested extarq."""

    # attach_handle
    @abc.abstractmethod
    def attach_handle_create(self, context, values):
        """Create a new attach_handle"""

    @abc.abstractmethod
    def attach_handle_get_by_uuid(self, context, uuid):
        """Get requested attach_handle"""

    @abc.abstractmethod
    def attach_handle_get_by_id(self, context, id):
        """Get requested attach_handle"""

    @abc.abstractmethod
    def attach_handle_get_by_filters(self, context,
                                     filters, sort_key='created_at',
                                     sort_dir='desc', limit=None,
                                     marker=None, columns_to_join=None):
        """Get requested deployable by filters."""

    @abc.abstractmethod
    def attach_handle_list(self, context):
        """Get requested list of attach_handles"""

    @abc.abstractmethod
    def attach_handle_delete(self, context, uuid):
        """Delete an attach_handle"""

    @abc.abstractmethod
    def attach_handle_update(self, context, uuid, values):
        """Update an attach_handle"""

    # control_path_id
    @abc.abstractmethod
    def control_path_create(self, context, values):
        """Create a new control path id"""

    @abc.abstractmethod
    def control_path_get_by_uuid(self, context, uuid):
        """Get requested control path id"""

    @abc.abstractmethod
    def control_path_get_by_filters(self, context,
                                    filters, sort_key='created_at',
                                    sort_dir='desc', limit=None,
                                    marker=None, columns_to_join=None):
        """Get requested deployable by filters."""

    @abc.abstractmethod
    def control_path_list(self, context):
        """Get requested list of control path ids"""

    @abc.abstractmethod
    def control_path_delete(self, context, uuid):
        """Delete a control path id"""

    @abc.abstractmethod
    def control_path_update(self, context, uuid, values):
        """Update a control path id"""



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from cyborg.common import rpc
from cyborg import version


def parse_args(argv, default_config_files=None):
    rpc.set_defaults(control_exchange='cyborg')
    version_string = version.version_info.release_string()
    cfg.CONF(argv[1:],
             project='cyborg',
             version=version_string,
             default_config_files=default_config_files)
    rpc.init(cfg.CONF)



# Copyright 2018 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import base as object_base

from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields

LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class Deployable(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # 1.0: Initial version
    # 1.1: Added rp_uuid, driver_name, bitstream_id, num_accel_in_use
    VERSION = '1.1'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.UUIDField(nullable=False),
        'parent_id': object_fields.IntegerField(nullable=True),
        # parent_id refers to the id of the deployable's parent node
        'root_id': object_fields.IntegerField(nullable=True),
        # root_id refers to the id of the deployable's root to for nested tree
        'name': object_fields.StringField(nullable=False),
        # name of the deployable
        'num_accelerators': object_fields.IntegerField(nullable=False),
        # number of accelerators spawned by this deployable
        'device_id': object_fields.IntegerField(nullable=False),
        # Foreign key constrain to reference device table
        'driver_name': object_fields.StringField(nullable=True),
        # Will change it to non-nullable after other driver report it.
        'rp_uuid': object_fields.UUIDField(nullable=True),
        # UUID of the Resource provider corresponding to this deployable
        'bitstream_id': object_fields.UUIDField(nullable=True),
    }

    def _get_parent_root_id(self, context):
        obj_dep = Deployable.get_by_id(context, self.parent_id)
        return obj_dep.root_id

    def create(self, context):
        """Create a Deployable record in the DB."""

        # FIXME: Add parent_uuid and root_uuid constrains when DB change to
        # parent_uuid & root_uuid

        values = self.obj_get_changes()
        db_dep = self.dbapi.deployable_create(context, values)
        self._from_db_object(self, db_dep)
        self.obj_reset_changes()

    @classmethod
    def get(cls, context, uuid):
        """Find a DB Deployable and return an Obj Deployable."""
        db_dep = cls.dbapi.deployable_get(context, uuid)
        obj_dep = cls._from_db_object(cls(context), db_dep)
        obj_dep.obj_reset_changes()
        return obj_dep

    @classmethod
    def get_by_id(cls, context, id):
        """Find a DB Deployable and return an Obj Deployable."""
        dpl_query = {"id": id}
        obj_dep = Deployable.get_by_filter(context, dpl_query)[0]
        obj_dep.obj_reset_changes()
        return obj_dep

    @classmethod
    def get_by_device_rp_uuid(cls, context, devrp_uuid):
        db_dep = cls.dbapi.deployable_get_by_rp_uuid(context, devrp_uuid)
        obj_dep = cls._from_db_object(cls(context), db_dep)
        return obj_dep

    @classmethod
    def list(cls, context, filters=None):
        """Return a list of Deployable objects."""
        if filters:
            sort_dir = filters.pop('sort_dir', 'desc')
            sort_key = filters.pop('sort_key', 'created_at')
            limit = filters.pop('limit', None)
            marker = filters.pop('marker_obj', None)
            db_deps = cls.dbapi.deployable_get_by_filters(context, filters,
                                                          sort_dir=sort_dir,
                                                          sort_key=sort_key,
                                                          limit=limit,
                                                          marker=marker)
        else:
            db_deps = cls.dbapi.deployable_list(context)
        obj_dpl_list = cls._from_db_object_list(db_deps, context)
        return obj_dpl_list

    def save(self, context):
        """Update a Deployable record in the DB."""
        updates = self.obj_get_changes()
        # TODO(Xinran): Will remove this if find some better way.
        updates.pop("uuid", None)
        updates.pop("created_at", None)
        if "updated_at" in updates.keys() and \
                updates["updated_at"] is not None:
            updates["updated_at"] = updates["updated_at"].replace(tzinfo=None)
        db_dep = self.dbapi.deployable_update(context, self.uuid, updates)
        self.obj_reset_changes()
        self._from_db_object(self, db_dep)

    def update(self, context, updates):
        """Update provided key, value pairs"""
        self.dbapi.deployable_update(context, self.uuid,
                                     updates)

    def destroy(self, context):
        """Delete a Deployable from the DB."""
        self.dbapi.deployable_delete(context, self.uuid)
        self.obj_reset_changes()

    @classmethod
    def get_by_filter(cls, context,
                      filters):
        obj_dpl_list = []
        db_dpl_list = cls.dbapi.deployable_get_by_filters(
            context,
            filters)

        if db_dpl_list:
            for db_dpl in db_dpl_list:
                obj_dpl = cls._from_db_object(cls(context), db_dpl)
                obj_dpl_list.append(obj_dpl)

        return obj_dpl_list

    @classmethod
    def get_list_by_device_id(cls, context, device_id):
        dep_filter = {'device_id': device_id}
        dep_obj_list = Deployable.list(context, dep_filter)
        return dep_obj_list

    @classmethod
    def get_by_name_deviceid(cls, context, name, device_id):
        dep_filter = {'name': name, 'device_id': device_id}
        dep_obj_list = Deployable.list(context, dep_filter)
        if len(dep_obj_list) != 0:
            return dep_obj_list[0]
        else:
            return None

    @classmethod
    def get_by_name(cls, context, name):
        dep_filter = {'name': name}
        dep_obj_list = Deployable.list(context, dep_filter)
        if len(dep_obj_list) != 0:
            return dep_obj_list[0]
        else:
            return None

    def get_cpid_list(self, context):
        query_filter = {"device_id": self.device_id}
        # TODO(Sundar) We should probably get cpid from objects layer,
        # not db layer
        cpid_list = self.dbapi.control_path_get_by_filters(
            context, query_filter)
        return cpid_list



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SPDK NVMFDRIVER module implementation.
"""

from oslo_log import log as logging

from cyborg.accelerator.common import exception
from cyborg.accelerator.drivers.spdk.spdk import SPDKDRIVER
from cyborg.accelerator.drivers.spdk.util import common_fun
from cyborg.accelerator.drivers.spdk.util.pyspdk.nvmf_client import NvmfTgt
from cyborg.accelerator.drivers.spdk.util.pyspdk.py_spdk import PySPDK

LOG = logging.getLogger(__name__)


class NVMFDRIVER(SPDKDRIVER):
    """NVMFDRIVER class.

    nvmf_tgt server app should be able to implement this driver.
    """

    SERVER = 'nvmf'

    def __init__(self, *args, **kwargs):
        super(NVMFDRIVER, self).__init__(*args, **kwargs)
        self.servers = common_fun.discover_servers()
        self.py = common_fun.get_py_client(self.SERVER)

    def discover_accelerator(self):
        if common_fun.check_for_setup_error(self.py, self.SERVER):
            return self.get_one_accelerator()

    def get_one_accelerator(self):
        acc_client = NvmfTgt(self.py)
        bdevs = acc_client.get_bdevs()
        # Display current blockdev list
        subsystems = acc_client.get_nvmf_subsystems()
        # Display nvmf subsystems
        accelerator_obj = {
            'server': self.SERVER,
            'bdevs': bdevs,
            'subsystems': subsystems
        }
        return accelerator_obj

    def install_accelerator(self, driver_id, driver_type):
        pass

    def uninstall_accelerator(self, driver_id, driver_type):
        pass

    def accelerator_list(self):
        return self.get_all_accelerators()

    def get_all_accelerators(self):
        accelerators = []
        for accelerator_i in range(len(self.servers)):
            accelerator = self.servers[accelerator_i]
            py_tmp = PySPDK(accelerator)
            if py_tmp.is_alive():
                accelerators.append(self.get_one_accelerator())
        return accelerators

    def update(self, driver_type, **kwargs):
        pass

    def attach_instance(self, instance_id):
        pass

    def detach_instance(self, instance_id):
        pass

    def delete_subsystem(self, nqn):
        """Delete a nvmf subsystem

        :param nqn: Target nqn(ASCII).
        :raise exception: Invaid
        """
        if nqn == "":
            acc_client = NvmfTgt(self.py)
            acc_client.delete_nvmf_subsystem(nqn)
        else:
            raise exception.Invalid('Delete nvmf subsystem failed.')

    def construct_subsystem(self,
                            nqn,
                            listen,
                            hosts,
                            serial_number,
                            namespaces
                            ):
        """Add a nvmf subsystem

        :param nqn: Target nqn(ASCII).
        :param listen: comma-separated list of Listen
        <trtype:transport_name traddr:address trsvcid:port_id>
        pairs enclosed in quotes.  Format:'trtype:transport0
        traddr:traddr0 trsvcid:trsvcid0,trtype:transport1
        traddr:traddr1 trsvcid:trsvcid1' etc.
        Example: 'trtype:RDMA traddr:192.168.100.8 trsvcid:4420,
        trtype:RDMA traddr:192.168.100.9 trsvcid:4420.'
        :param hosts: Whitespace-separated list of host nqn list.
        :param serial_number: Example: 'SPDK00000000000001.
        :param namespaces: Whitespace-separated list of namespaces.
        :raise exception: Invaid
        """
        if ((namespaces != '' and listen != '') and
                (hosts != '' and serial_number != '')) and nqn != '':
            acc_client = NvmfTgt(self.py)
            acc_client.construct_nvmf_subsystem(nqn,
                                                listen,
                                                hosts,
                                                serial_number,
                                                namespaces
                                                )
        else:
            raise exception.Invalid('Construct nvmf subsystem failed.')



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import alembic
from alembic import config as alembic_config
import alembic.migration as alembic_migration
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade

from cyborg.db.sqlalchemy import models


def _alembic_config():
    path = os.path.join(os.path.dirname(__file__), 'alembic.ini')
    config = alembic_config.Config(path)
    return config


def version(config=None, engine=None):
    """Current database version.

    :returns: Database version
    :rtype: string
    """
    if engine is None:
        engine = enginefacade.get_legacy_facade().get_engine()
    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        return context.get_current_revision()


def upgrade(revision, config=None):
    """Used for upgrading database.

    :param version: Desired database version
    :type version: string
    """
    revision = revision or 'head'
    config = config or _alembic_config()
    alembic.command.upgrade(config, revision)


def create_schema(config=None, engine=None):
    """Create database schema from models description.

    Can be used for initial installation instead of upgrade('head').
    """
    if engine is None:
        engine = enginefacade.get_legacy_facade().get_engine()

    if version(engine=engine) is not None:
        raise db_exc.DBMigrationError("DB schema is already under version"
                                      " control. Use upgrade() instead")

    models.Base.metadata.create_all(engine)
    stamp('head', config=config)


def downgrade(revision, config=None):
    """Used for downgrading database.

    :param version: Desired database version
    :type version: string
    """
    revision = revision or 'base'
    config = config or _alembic_config()
    return alembic.command.downgrade(config, revision)


def stamp(revision, config=None):
    """Stamps database with provided revision.

    Don't run any migrations.

    :param revision: Should match one from repository or head - to stamp
                     database with most recent revision
    :type revision: string
    """
    config = config or _alembic_config()
    return alembic.command.stamp(config, revision=revision)


def revision(message=None, autogenerate=False, config=None):
    """Creates template for migration.

    :param message: Text that will be used for migration title
    :type message: string
    :param autogenerate: If True - generates diff based on current database
                         state
    :type autogenerate: bool
    """
    config = config or _alembic_config()
    return alembic.command.revision(config, message=message,
                                    autogenerate=autogenerate)



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
import unittest
from unittest import mock

from oslo_serialization import jsonutils

from cyborg.api.controllers import base
from cyborg.api.controllers.v2 import arqs
from cyborg.common import exception
from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import fake_device_profile
from cyborg.tests.unit import fake_extarq


class TestARQsController(v2_test.APITestV2):

    ARQ_URL = '/accelerator_requests'

    def setUp(self):
        super(TestARQsController, self).setUp()
        self.headers = self.gen_headers(self.context)
        self.fake_extarqs = fake_extarq.get_fake_extarq_objs()
        self.fake_bind_extarqs = fake_extarq.get_fake_extarq_bind_objs()
        self.fake_resolved_extarqs = (
            fake_extarq.get_fake_extarq_resolved_objs())
        self.arqs_controller = arqs.ARQsController()

    def _validate_links(self, links, arq_uuid):
        has_self_link = False
        for link in links:
            if link['rel'] == 'self':
                has_self_link = True
                url = link['href']
                components = url.split('/')
                self.assertEqual(components[-1], arq_uuid)
        self.assertTrue(has_self_link)

    def _validate_arq(self, in_arq, out_arq):
        for field in in_arq.keys():
            if field != 'id':
                self.assertEqual(in_arq[field], out_arq[field])

        # Check that the link is properly set up
        self._validate_links(out_arq['links'], in_arq['uuid'])

    @mock.patch('cyborg.objects.ExtARQ.get')
    def test_get_one_by_uuid(self, mock_extarq):
        in_extarq = self.fake_extarqs[0]
        in_arq = in_extarq.arq
        mock_extarq.return_value = in_extarq
        uuid = in_arq['uuid']

        url = self.ARQ_URL + '/%s'
        out_arq = self.get_json(url % uuid, headers=self.headers)
        mock_extarq.assert_called_once()
        self._validate_arq(in_arq, out_arq)

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all(self, mock_extarqs):
        # test get_all of any bind_state
        mock_extarqs.return_value = self.fake_extarqs
        data = self.get_json(self.ARQ_URL, headers=self.headers)
        out_arqs = data['arqs']

        result = isinstance(out_arqs, list)
        self.assertTrue(result)
        self.assertEqual(len(out_arqs), len(self.fake_extarqs))
        for in_extarq, out_arq in zip(self.fake_extarqs, out_arqs):
            self._validate_arq(in_extarq.arq, out_arq)

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all_with_instance(self, mock_extarqs):
        # test get_all with instance
        mock_extarqs.return_value = self.fake_bind_extarqs
        instance_uuid = self.fake_bind_extarqs[0].arq.instance_uuid
        url = '%s?instance=%s' % (self.ARQ_URL, instance_uuid)
        data = self.get_json(url, headers=self.headers)
        out_arqs = data['arqs']

        result = isinstance(out_arqs, list)
        self.assertTrue(result)
        self.assertEqual(len(out_arqs), len(self.fake_bind_extarqs[:3]))
        for in_extarq, out_arq in zip(self.fake_bind_extarqs[:2], out_arqs):
            self._validate_arq(in_extarq.arq, out_arq)

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all_with_bind_state(self, mock_extarqs):
        # test get_all with valid bind_state(resolved)
        mock_extarqs.return_value = self.fake_resolved_extarqs
        url = '%s?bind_state=resolved' % self.ARQ_URL
        data = self.get_json(url, headers=self.headers)
        out_arqs = data['arqs']

        result = isinstance(out_arqs, list)
        self.assertTrue(result)
        self.assertEqual(len(out_arqs), len(self.fake_resolved_extarqs[1:]))
        for in_extarq, out_arq in zip(self.fake_resolved_extarqs[1:],
                                      out_arqs):
            self._validate_arq(in_extarq.arq, out_arq)

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all_with_instance_and_bind_state(self, mock_extarqs):
        # test get_all with instance and valid bind_state(resolved)
        mock_extarqs.return_value = self.fake_bind_extarqs[:3]
        instance_uuid = self.fake_bind_extarqs[0].arq.instance_uuid
        url = '%s?instance=%s&bind_state=resolved' % (
            self.ARQ_URL, instance_uuid)
        data = self.get_json(url, headers=self.headers)
        out_arqs = data['arqs']

        result = isinstance(out_arqs, list)
        self.assertTrue(result)
        self.assertEqual(len(out_arqs), len(self.fake_bind_extarqs[:2]))
        for in_extarq, out_arq in zip(self.fake_bind_extarqs[:2], out_arqs):
            self._validate_arq(in_extarq.arq, out_arq)

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all_with_http_client_LOCKED(self, mock_extarqs):
        # test get_all if not all ARQs are in bound state
        mock_extarqs.return_value = self.fake_bind_extarqs
        instance_uuid = self.fake_bind_extarqs[0].arq.instance_uuid
        url = '%s?instance=%s&bind_state=resolved' % (
            self.ARQ_URL, instance_uuid)
        try:
            self.get_json(url, headers=self.headers)
        except Exception as e:
            exc = e
        self.assertIn('423 Locked', exc.args[0])

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all_with_invalid_bind_state(self, mock_extarqs):
        # test get_all with bind_state=started
        mock_extarqs.return_value = self.fake_extarqs
        instance_uuid = self.fake_extarqs[0].arq.instance_uuid
        url = '%s?instance=%s&bind_state=started' % (
            self.ARQ_URL, instance_uuid)
        exc = None
        try:
            self.get_json(url, headers=self.headers)
        except Exception as e:
            exc = e
        # TODO(all) Cyborg does not have fake HTTPRequest Object now, so just
        # use assertIn here, improve this case with assertRaises later.
        self.assertIn(
            "Bad state: started for ARQ: None. Expected state(s): "
            "[\\\'resolved\\\']", exc.args[0])

        url = '%s?bind_state=started' % (self.ARQ_URL)
        exc = None
        try:
            self.get_json(url, headers=self.headers)
        except Exception as e:
            exc = e
        # TODO(all) Cyborg does not have fake HTTPRequest Object now, so just
        # use assertIn here, improve this case with assertRaises later.
        self.assertIn(
            "Bad state: started for ARQ: None. Expected state(s): "
            "[\\\'resolved\\\']", exc.args[0])

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_get_all_with_invalid_arq_state(self, mock_extarqs):
        # test get_all response "423 Locked"
        # set ARQ state to 'BindStarted'
        self.fake_extarqs[0].arq.state = 'BindStarted'
        mock_extarqs.return_value = self.fake_extarqs
        instance_uuid = self.fake_extarqs[0].arq.instance_uuid
        url = '%s?instance=%s&bind_state=resolved' % (
            self.ARQ_URL, instance_uuid)
        response = self.get_json(url, headers=self.headers, expect_errors=True)
        self.assertEqual(HTTPStatus.LOCKED, response.status_int)

    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.arq_create')
    def test_create(self, mock_obj_extarq, mock_obj_dp):
        dp_list = fake_device_profile.get_obj_devprofs()
        mock_obj_dp.return_value = dp = dp_list[0]
        mock_obj_extarq.side_effect = self.fake_extarqs
        params = {'device_profile_name': dp['name']}
        response = self.post_json(self.ARQ_URL, params, headers=self.headers)
        data = jsonutils.loads(response.__dict__['controller_output'])
        out_arqs = data['arqs']

        self.assertEqual(HTTPStatus.CREATED, response.status_int)
        self.assertEqual(len(out_arqs), 3)
        for in_extarq, out_arq in zip(self.fake_extarqs, out_arqs):
            self._validate_arq(in_extarq.arq, out_arq)
        for idx, out_arq in enumerate(out_arqs):
            dp_group_id = idx
            self.assertEqual(dp_group_id, out_arq['device_profile_group_id'])

    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.arq_create')
    def test_create_with_xilinx_fpga(self, mock_obj_extarq, mock_obj_dp):
        xilinx_fpga_dp = fake_device_profile.get_xilinx_fpga_devprof()
        mock_obj_dp.return_value = dp = xilinx_fpga_dp
        fake_xilinx_fpga_objs = fake_extarq.get_fake_xilinx_fpga_extarq_objs()
        mock_obj_extarq.side_effect = fake_xilinx_fpga_objs
        params = {'device_profile_name': dp['name']}
        response = self.post_json(self.ARQ_URL, params, headers=self.headers)
        data = jsonutils.loads(response.__dict__['controller_output'])
        out_arqs = data['arqs']

        self.assertEqual(HTTPStatus.CREATED, response.status_int)
        self.assertEqual(len(out_arqs), 2)
        for in_extarq, out_arq in zip(fake_xilinx_fpga_objs, out_arqs):
            self._validate_arq(in_extarq.arq, out_arq)
        for idx, out_arq in enumerate(out_arqs):
            dp_group_id = idx
            self.assertEqual(dp_group_id, out_arq['device_profile_group_id'])

    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    @mock.patch('cyborg.objects.ExtARQ.create')
    def test_create_with_wrong_dp(self, mock_obj_extarq, mock_obj_dp):
        params = {'device_profile_name': 'wrong_device_profile_name'}
        mock_obj_dp.side_effect = exception.ResourceNotFound(
            resource='Device Profile',
            msg='with name=%s' % params.get('device_profile_name'))
        mock_obj_extarq.side_effect = self.fake_extarqs
        exc = None
        try:
            self.post_json(self.ARQ_URL, params, headers=self.headers)
        except Exception as e:
            exc = e
        self.assertIn(
            "Device Profile not found with "
            "name=wrong_device_profile_name", exc.args[0])

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.arq_delete_by_uuid')
    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.'
                'arq_delete_by_instance_uuid')
    def test_delete(self, mock_by_inst, mock_by_arq):
        url = self.ARQ_URL
        arq = self.fake_extarqs[0].arq
        instance = arq.instance_uuid

        mock_by_arq.return_value = None
        args = '?' + "arqs=" + str(arq['uuid'])
        response = self.delete(url + args, headers=self.headers)
        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)

        mock_by_inst.return_value = None
        args = '?' + "instance=" + instance
        response = self.delete(url + args, headers=self.headers)
        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)

    @unittest.skip("Need more code to implement _get_resource in rbac")
    def test_delete_with_non_default(self):
        value = {"is_admin": False, "roles": "user", "is_admin_project": False}
        ct = self.gen_context(value)
        headers = self.gen_headers(ct)
        url = self.ARQ_URL
        arq = self.fake_extarqs[0].arq
        args = '?' + "arqs=" + str(arq['uuid'])
        exc = None
        try:
            self.delete(url + args, headers=headers)
        except Exception as e:
            exc = e
        # Cyborg does not raise different exception when policy check failed
        # now, improve this case with assertRaises later.
        self.assertIn("Bad response: 403 Forbidden", exc.args[0])

    @mock.patch.object(arqs.ARQsController, '_check_if_already_bound')
    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.arq_apply_patch')
    def test_apply_patch(self, mock_apply_patch, mock_check_if_bound):
        """Test the happy path."""
        patch_list, device_rp_uuid = fake_extarq.get_patch_list()
        arq_uuids = list(patch_list.keys())
        obj_extarq = self.fake_extarqs[0]
        valid_fields = {
            arq_uuid: {
                'hostname': obj_extarq.arq.hostname,
                'device_rp_uuid': device_rp_uuid,
                'instance_uuid': obj_extarq.arq.instance_uuid}
            for arq_uuid in arq_uuids}

        self.patch_json(self.ARQ_URL, params=patch_list,
                        headers=self.headers)

        mock_apply_patch.assert_called_once_with(mock.ANY, patch_list,
                                                 valid_fields)
        mock_check_if_bound.assert_called_once_with(mock.ANY, valid_fields)

    @mock.patch.object(arqs.ARQsController, '_check_if_already_bound')
    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.arq_apply_patch')
    def test_apply_patch_allow_project_id(
            self, mock_apply_patch, mock_check_if_bound):
        patch_list, _ = fake_extarq.get_patch_list()
        for arq_uuid, patch in patch_list.items():
            patch.append({'path': '/project_id', 'op': 'add',
                          'value': 'b1c76756ac2e482789a8e1c5f4bf065e'})
        arq_uuids = list(patch_list.keys())
        valid_fields = {
            arq_uuid: {
                'hostname': 'myhost',
                'device_rp_uuid': 'fb16c293-5739-4c84-8590-926f9ab16669',
                'instance_uuid': '5922a70f-1e06-4cfd-88dd-a332120d7144',
                'project_id': 'b1c76756ac2e482789a8e1c5f4bf065e'}
            for arq_uuid in arq_uuids}

        self.patch_json(self.ARQ_URL, params=patch_list,
                        headers={base.Version.current_api_version:
                                 '2.1'})
        mock_apply_patch.assert_called_once_with(mock.ANY, patch_list,
                                                 valid_fields)
        mock_check_if_bound.assert_called_once_with(mock.ANY, valid_fields)

    def test_apply_patch_not_allow_project_id(self):
        patch_list, _ = fake_extarq.get_patch_list()
        for arq_uuid, patch in patch_list.items():
            patch.append({'path': '/project_id', 'op': 'add',
                          'value': 'b1c76756ac2e482789a8e1c5f4bf065e'})
        response = self.patch_json(self.ARQ_URL, params=patch_list,
                                   headers=self.headers,
                                   expect_errors=True)
        self.assertEqual(HTTPStatus.NOT_ACCEPTABLE, response.status_code)
        self.assertTrue(response.json['error_message'])

    # TODO(all): Add exception test cases for apply_patch.

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_check_if_bound(self, mock_extarq_list):
        """Test the happy path."""
        extarqs = fake_extarq.get_fake_extarq_objs()
        mock_extarq_list.return_value = extarqs

        # Not the instance UUID in extarqs above
        instance_uuid = 'ffbb66f6-99f6-4a85-a90c-fd8e8fb35f16'
        valid_fields = {
            extarq.arq['uuid']: {
                'hostname': 'myhost',
                'device_rp_uuid': 'fb16c293-5739-4c84-8590-926f9ab16669',
                'instance_uuid': instance_uuid}
            for extarq in extarqs}

        self.arqs_controller._check_if_already_bound(
            self.context, valid_fields)
        mock_extarq_list.assert_called_once_with(self.context)

    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_check_if_bound_exception(self, mock_extarq_list):
        """Test that an exception is raised if binding request specifies
           an instance that already has ARQs.
        """
        extarqs = fake_extarq.get_fake_extarq_objs()
        mock_extarq_list.return_value = extarqs

        # Same instance UUID as in extarqs above, thus triggering exception
        instance_uuid = extarqs[0].arq['instance_uuid']
        valid_fields = {
            extarq.arq['uuid']: {
                'hostname': 'myhost',
                'device_rp_uuid': 'fb16c293-5739-4c84-8590-926f9ab16669',
                'instance_uuid': instance_uuid}
            for extarq in extarqs}

        expected_err = ('Instance %s already has accelerator requests. '
                        'Cannot bind additional ARQs.') % instance_uuid

        self.assertRaisesRegex(
            exception.PatchError, expected_err,
            self.arqs_controller._check_if_already_bound,
            self.context, valid_fields)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils
from oslo_log import log
import oslo_messaging as messaging
from oslo_service import service
from oslo_service import wsgi
from oslo_utils import importutils

from cyborg.api import app
from cyborg.common import config
from cyborg.common import exception
from cyborg.common.i18n import _
from cyborg.common import rpc
from cyborg.conf import CONF
from cyborg import context
from cyborg import objects
from cyborg.objects import base as objects_base


LOG = log.getLogger(__name__)


class RPCService(service.Service):
    def __init__(self, manager_module, manager_class, topic, host=None):
        super(RPCService, self).__init__()
        self.topic = topic
        self.host = host or CONF.host
        manager_module = importutils.try_import(manager_module)
        manager_class = getattr(manager_module, manager_class)
        self.manager = manager_class(self.topic, self.host)
        self.rpcserver = None

    def start(self):
        super(RPCService, self).start()

        target = messaging.Target(topic=self.topic, server=self.host)
        endpoints = [self.manager]
        serializer = objects_base.CyborgObjectSerializer()
        self.rpcserver = rpc.get_server(target, endpoints, serializer)
        self.rpcserver.start()

        admin_context = context.get_admin_context()
        self.tg.add_dynamic_timer(
            self.manager.periodic_tasks,
            periodic_interval_max=CONF.periodic_interval,
            context=admin_context)

        LOG.info('Created RPC server for service %(service)s on host '
                 '%(host)s.',
                 {'service': self.topic, 'host': self.host})

    def stop(self, graceful=True):
        try:
            self.rpcserver.stop()
            self.rpcserver.wait()
        except Exception as e:
            LOG.exception('Service error occurred when stopping the '
                          'RPC server. Error: %s', e)

        super(RPCService, self).stop(graceful=graceful)
        LOG.info('Stopped RPC server for service %(service)s on host '
                 '%(host)s.',
                 {'service': self.topic, 'host': self.host})


def prepare_service(argv=None):
    log.register_options(CONF)
    log.set_defaults(default_log_levels=CONF.default_log_levels)

    argv = argv or []
    config.parse_args(argv)

    log.setup(CONF, 'cyborg')
    objects.register_all()


def process_launcher():
    return service.ProcessLauncher(CONF, restart_method='mutate')


class WSGIService(service.ServiceBase):
    """Provides ability to launch cyborg API from wsgi app."""

    def __init__(self, name, use_ssl=False):
        """Initialize, but do not start the WSGI server.

        :param name: The name of the WSGI server given to the loader.
        :param use_ssl: Wraps the socket in an SSL context if True.
        :returns: None
        """
        self.name = name
        self.app = app.load_app()
        self.workers = (CONF.api.api_workers or
                        processutils.get_worker_count())
        if self.workers and self.workers < 1:
            raise exception.ConfigInvalid(
                _("api_workers value of %d is invalid, "
                  "must be greater than 0.") % self.workers)

        self.server = wsgi.Server(CONF, self.name, self.app,
                                  host=CONF.api.host_ip,
                                  port=CONF.api.port,
                                  use_ssl=use_ssl)

    def start(self):
        """Start serving this service using loaded configuration.

        :returns: None
        """
        self.server.start()

    def stop(self):
        """Stop serving this API.

        :returns: None
        """
        self.server.stop()

    def wait(self):
        """Wait for the service to stop serving this API.

        :returns: None
        """
        self.server.wait()

    def reset(self):
        """Reset server greenpool size to default.

        :returns: None
        """
        self.server.reset()



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cyborg base exception handling.

SHOULD include dedicated exception logging.

"""

from http import HTTPStatus
from oslo_log import log

from cyborg.common.i18n import _
from cyborg.conf import CONF


LOG = log.getLogger(__name__)


class CyborgException(Exception):
    """Base Cyborg Exception

    To correctly use this class, inherit from it and define
    a '_msg_fmt' property. That message will get printf'd
    with the keyword arguments provided to the constructor.

    If you need to access the message from an exception you should use
    str(exc).
    """
    _msg_fmt = _("An unknown exception occurred.")
    code = HTTPStatus.INTERNAL_SERVER_ERROR
    headers = {}
    safe = False

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            try:
                message = self._msg_fmt % kwargs
            except Exception:
                # kwargs doesn't match a variable in self._msg_fmt
                # log the issue and the kwargs
                LOG.exception('Exception in string format operation')
                for name, value in kwargs.items():
                    LOG.error("%(name)s: %(value)s",
                              {"name": name, "value": value})

                if CONF.fatal_exception_format_errors:
                    raise
                else:
                    # at least get the core self._msg_fmt out if something
                    # happened
                    message = self._msg_fmt

        super(CyborgException, self).__init__(message)

    def __str__(self):
        """Encode to utf-8 then wsme api can consume it as well."""
        return self.args[0]


class Forbidden(CyborgException):
    _msg_fmt = _("Forbidden")
    code = HTTPStatus.FORBIDDEN


class ARQBadState(CyborgException):
    _msg_fmt = _('Bad state: %(state)s for ARQ: %(uuid)s. '
                 'Expected state(s): %(expected)s')


class AttachHandleAlreadyExists(CyborgException):
    _msg_fmt = _("AttachHandle with uuid %(uuid)s already exists.")


class AttachHandleUUIDNeeded(CyborgException):
    _msg_fmt = _("Need to provide AttachHandle uuid.")


class ControlpathIDAlreadyExists(CyborgException):
    _msg_fmt = _("ControlpathID with uuid %(uuid)s already exists.")


class ConfigInvalid(CyborgException):
    _msg_fmt = _("Invalid configuration file. %(error_msg)s")


class DeviceAlreadyExists(CyborgException):
    _msg_fmt = _("Device with uuid %(uuid)s already exists.")


class DeviceProfileAlreadyExists(CyborgException):
    _msg_fmt = _("DeviceProfile with uuid %(uuid)s already exists.")


class DeviceProfileNameNeeded(CyborgException):
    _msg_fmt = _("DeviceProfile name needed.")


class DeviceProfileGroupsExpected(CyborgException):
    _msg_fmt = _("DeviceProfile needs groups field.")


class DeployableAlreadyExists(CyborgException):
    _msg_fmt = _("Deployable with uuid %(uuid)s already exists.")


class ExtArqAlreadyExists(CyborgException):
    _msg_fmt = _("ExtArq with uuid %(uuid)s already exists.")


class ExpectedOneObject(CyborgException):
    _msg_fmt = _("Expected one object of type %(obj)s "
                 "but got %(count)s.")


class InUse(CyborgException):
    _msg_fmt = _("%(object) with %(ident)s is in use.")


class Invalid(CyborgException):
    _msg_fmt = _("Invalid parameters.")
    code = HTTPStatus.BAD_REQUEST


class InvalidIdentity(Invalid):
    _msg_fmt = _("Expected a uuid/id but received %(identity)s.")


class InvalidUUID(Invalid):
    _msg_fmt = _("Expected a uuid but received %(uuid)s.")


class InvalidJsonType(Invalid):
    _msg_fmt = _("%(value)s is not JSON serializable.")


class InvalidAPIVersionString(Invalid):
    _msg_fmt = _("API Version String %(version)s is of invalid format. Must "
                 "be of format MajorNum.MinorNum.")


# TODO(All): Consider whether Placement/Image exceptions can be included here.
class InvalidAPIResponse(Invalid):
    _msg_fmt = _('Bad API response from %(service)s for %(api)s API. '
                 'Details: %(msg)s')


# Cannot be templated as the error syntax varies.
# msg needs to be constructed when raised.
class InvalidParameterValue(Invalid):
    _msg_fmt = _("%(err)s")


class NeedAtleastOne(Invalid):
    _msg_fmt = _("Need at least one %(obj)s.")


class PatchError(Invalid):
    _msg_fmt = _("Couldn't apply patch. Reason: %(reason)s")


class NotAuthorized(CyborgException):
    _msg_fmt = _("Not authorized.")
    code = HTTPStatus.UNAUTHORIZED


class HTTPForbidden(Forbidden):
    _msg_fmt = _("Access was denied to the following resource: %(resource)s")


class NotFound(CyborgException):
    _msg_fmt = _("Resource could not be found.")
    code = HTTPStatus.NOT_FOUND


class ServiceUnavailable(CyborgException):
    _msg_fmt = _("Service is unavailable at this time.")
    code = HTTPStatus.SERVICE_UNAVAILABLE


class ServiceNotFound(NotFound):
    _msg_fmt = _("Service %(service_id)s could not be found.")


class ConfGroupForServiceTypeNotFound(ServiceNotFound):
    _msg_fmt = _("No conf group name could be found for service type "
                 "%(stype)s.")


class InvalidDeployType(CyborgException):
    _msg_fmt = _("Deployable have an invalid type")


class Conflict(CyborgException):
    _msg_fmt = _('Conflict.')
    code = HTTPStatus.CONFLICT


class DuplicateDeviceName(Conflict):
    _msg_fmt = _("A device with name %(name)s already exists.")


class DuplicateDeviceProfileName(Conflict):
    _msg_fmt = _("A device_profile with name %(name)s already exists.")


class DuplicateDeployableName(Conflict):
    _msg_fmt = _("A deployable with name %(name)s already exists.")


class PlacementEndpointNotFound(NotFound):
    _msg_fmt = _("Placement API endpoint not found")


class PlacementResourceProviderNotFound(NotFound):
    _msg_fmt = _("Placement resource provider not found: "
                 "%(resource_provider)s.")


class PlacementInventoryNotFound(NotFound):
    _msg_fmt = _("Placement inventory not found for resource provider "
                 "%(resource_provider)s, resource class %(resource_class)s.")


class PlacementInventoryUpdateConflict(Conflict):
    _msg_fmt = _("Placement inventory update conflict for resource provider "
                 "%(resource_provider)s, resource class %(resource_class)s.")


class ObjectActionError(CyborgException):
    _msg_fmt = _('Object action %(action)s failed because: %(reason)s')


class AttributeInvalid(CyborgException):
    _msg_fmt = _("Attribute is invalid")


class AttributeAlreadyExists(CyborgException):
    _msg_fmt = _("Attribute with uuid %(uuid)s already exists.")


# An exception with this name is used on both sides of the placement/
# cyborg interaction.
class ResourceClassNotFound(NotFound):
    _msg_fmt = _("No such resource class %(name_or_uuid)s.")


class ResourceProviderInUse(CyborgException):
    _msg_fmt = _("Resource provider has allocations.")


class ResourceProviderRetrievalFailed(CyborgException):
    _msg_fmt = _("Failed to get resource provider with UUID %(uuid)s")


class ResourceProviderAggregateRetrievalFailed(CyborgException):
    _msg_fmt = _("Failed to get aggregates for resource provider with UUID"
                 " %(uuid)s")


class ResourceProviderTraitRetrievalFailed(CyborgException):
    _msg_fmt = _("Failed to get traits for resource provider with UUID"
                 " %(uuid)s")


class ResourceProviderCreationFailed(CyborgException):
    _msg_fmt = _("Failed to create resource provider %(name)s")


class ResourceProviderDeletionFailed(CyborgException):
    _msg_fmt = _("Failed to delete resource provider %(uuid)s")


class ResourceProviderUpdateFailed(CyborgException):
    _msg_fmt = _("Failed to update resource provider via URL %(url)s: "
                 "%(error)s")


class ResourceProviderSyncFailed(CyborgException):
    _msg_fmt = _("Failed to synchronize the placement service with resource "
                 "provider information supplied by the compute host.")


class PlacementAPIConnectFailure(CyborgException):
    _msg_fmt = _("Unable to communicate with the Placement API.")


class PlacementServerError(CyborgException):
    _msg_fmt = _("Placement Server has some error at this time.")


class PlacementAPIConflict(CyborgException):
    """Any 409 error from placement APIs should use (a subclass of) this
    exception.
    """
    _msg_fmt = _("A conflict was encountered attempting to invoke the "
                 "placement API at URL %(url)s: %(error)s")


class ResourceProviderUpdateConflict(PlacementAPIConflict):
    """A 409 caused by generation mismatch from attempting to update an
    existing provider record or its associated data (aggregates, traits, etc.).
    """
    _msg_fmt = _("A conflict was encountered attempting to update resource "
                 "provider %(uuid)s (generation %(generation)d): %(error)s")


class TraitCreationFailed(CyborgException):
    _msg_fmt = _("Failed to create trait %(name)s: %(error)s")


class TraitRetrievalFailed(CyborgException):
    _msg_fmt = _("Failed to retrieve traits from the placement API: %(error)s")


class InvalidResourceClass(Invalid):
    _msg_fmt = _("Resource class '%(resource_class)s' invalid.")


class InvalidResourceAmount(Invalid):
    _msg_fmt = _("Resource amounts must be integers. Received '%(amount)s'.")


class InvalidInventory(Invalid):
    _msg_fmt = _("Inventory for '%(resource_class)s' on "
                 "resource provider '%(resource_provider)s' invalid.")


# An exception with this name is used on both sides of the placement/
# cyborg interaction.
class InventoryInUse(InvalidInventory):
    # NOTE(mriedem): This message cannot change without impacting the
    # cyborg.services.client.report._RE_INV_IN_USE regex.
    _msg_fmt = _("Inventory for '%(resource_classes)s' on "
                 "resource provider '%(resource_provider)s' in use.")


class QuotaNotFound(NotFound):
    _msg_fmt = _("Quota could not be found")


class QuotaUsageNotFound(QuotaNotFound):
    _msg_fmt = _("Quota usage for project %(project_id)s could not be found.")


class QuotaResourceUnknown(QuotaNotFound):
    _msg_fmt = _("Unknown quota resources %(unknown)s.")


class InvalidReservationExpiration(Invalid):
    _msg_fmt = _("Invalid reservation expiration %(expire)s.")


class GlanceConnectionFailed(CyborgException):
    _msg_fmt = _("Connection to glance host %(server)s failed: "
                 "%(reason)s")


class ImageUnacceptable(Invalid):
    _msg_fmt = _("Image %(image_id)s is unacceptable: %(reason)s")


class ImageNotAuthorized(CyborgException):
    _msg_fmt = _("Not authorized for image %(image_id)s.")


class ImageBadRequest(Invalid):
    _msg_fmt = _("Request of image %(image_id)s got BadRequest response: "
                 "%(response)s")


class InvalidDriver(Invalid):
    _msg_fmt = _("Found an invalid driver: %(name)s")


class InvalidVGPUType(Invalid):
    _msg_fmt = _("Invalid requested vGPU type: %(name)s")


class InvalidvGPUConfig(Invalid):
    _msg_fmt = _("Invalid vGPU config: %(reason)s")


class PciDeviceWrongAddressFormat(Invalid):
    _msg_fmt = _("The PCI address %(address)s has an incorrect format.")


class InvalidType(Invalid):
    _msg_fmt = _("Invalid type for %(obj)s: %(type)s."
                 "Expected: %(expected)s")


class ResourceNotFound(NotFound):
    _msg_fmt = _("%(resource)s not found %(msg)s")


class NotAcceptable(CyborgException):
    _msg_fmt = _("Request not acceptable.")
    code = HTTPStatus.NOT_ACCEPTABLE


class FPGAProgramError(CyborgException):
    _msg_fmt = _("FPGA programming failed with return %(ret)s.")


class PciDeviceNotFoundById(NotFound):
    _msg_fmt = _("PCI device %(id)s not found")



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonpatch
import wsme


from cyborg.common.i18n import _


JSONPATCH_EXCEPTIONS = (jsonpatch.JsonPatchException,
                        jsonpatch.JsonPointerException,
                        KeyError)


def apply_jsonpatch(doc, patch):
    for p in patch:
        if p['op'] == 'add' and p['path'].count('/') == 1:
            if p['path'].lstrip('/') not in doc:
                msg = _('Adding a new attribute (%s) to the root of '
                        ' the resource is not allowed')
                raise wsme.exc.ClientSideError(msg % p['path'])
    return jsonpatch.apply_patch(doc, jsonpatch.JsonPatch(patch))
